import os
import pytest
import torch as th
from vidrial.jit import settings, PickBest, PickAny
from vidrial.kernels.sympow.dimensions import sympow_dim
from vidrial.py_utils.test_utils import diff
from vidrial.kernels.sympow.interface import interface as sympow
from vidrial.kernels.sympow_mma.op import op as sympow_mma

# b, t, d, p, d_tile, duplicate_correction
ALL_PROBLEMS = [
    (b, t, d, e, p, d_tile, True, expand_dim, dtype)
    for d, e, p, d_tile, dtype in [(64, 64, 2, 8, th.bfloat16)]
    for b in [16, 8, 4, 2, 1, 32, 64, 128]
    for t in [1024, 4096]
    for expand_dim in [-1, -2]
]

# ---- Check sympow_mma == sympow_expand + mma on GPU ----
@pytest.mark.parametrize("b, t, d, e, p, d_tile, duplicate_correction, expand_dim, dtype", ALL_PROBLEMS, ids=str)
@pytest.mark.skipif(
    "not config.getoption('--longrun')",
    reason="Only run when --longrun is given",
)
def test_sympow_expandA_mma(b, t, d, e, p, d_tile, duplicate_correction, expand_dim, dtype):
    D = sympow_dim(d, p, d_tile)
    release_mode = os.environ.get("JIT_PACKAGE_MODE", "0") == "1"
    if expand_dim == -2: # expand M
        A = th.randn(b, d, t, dtype=dtype, device="cuda", requires_grad=True) * 1e-3
        B = th.randn(b, t, e, dtype=dtype, device="cuda", requires_grad=True) * 1e-3
        A_ref = A.detach().clone().requires_grad_(True)
        B_ref = B.detach().clone().requires_grad_(True)
    else: # expand K
        A = th.randn(b, t, d, dtype=dtype, device="cuda", requires_grad=True) * 1e-3
        B = th.randn(b, D, e, dtype=dtype, device="cuda", requires_grad=True) * 1e-3
        A_ref = A.detach().clone().requires_grad_(True)
        B_ref = B.detach().clone().requires_grad_(True)

    for t in [A, B, A_ref, B_ref]:
        t.retain_grad()

    with settings.set(policy=PickBest if release_mode else PickAny):
        C = sympow_mma(A, B, expand_dim, p, d_tile, duplicate_correction)
    assert C.shape == (b, D, e) if expand_dim == -2 else (b, t, e)
    eA = sympow(A_ref, p, d_tile, expand_dim, duplicate_correction)
    C_ref = eA @ B_ref
    diff(C, C_ref, verbose=True, atol=1e-1, rtol=1e-2)


    th.autograd.backward(C_ref, th.ones_like(C_ref))
    with settings.set(policy=PickBest if release_mode else PickAny):
        th.autograd.backward(C, th.ones_like(C))

    diff(A.grad, A_ref.grad, verbose=True, atol=1e-1, rtol=1e-2)
    diff(B.grad, B_ref.grad, verbose=True, atol=1e-1, rtol=1e-2)
    



import ctypes
import os
import logging
from functools import lru_cache

logger = logging.getLogger(__name__)

from vidrial.jit.ctypes import map_ctype


class Runtime:
    def __init__(self, path: str, entrypoint: str = 'launch') -> None:
        self.path = path
        self.entrypoint = entrypoint
        self.lib = None

        assert Runtime.is_path_valid(self.path)

    @staticmethod
    def is_path_valid(path: str) -> bool:
        # Exists and is a directory
        if not os.path.exists(path) or not os.path.isdir(path):
            return False

        # Contains all necessary files
        files = ['kernel.cu', 'kernel.so']
        return all(os.path.exists(os.path.join(path, file)) for file in files)

    def __call__(self, *args) -> int:
        # Load SO file
        if self.lib is None:
            self.lib = ctypes.CDLL(os.path.join(self.path, 'kernel.so'))

        cargs = [map_ctype(arg) for arg in args]
        ret = self.lib[self.entrypoint](*cargs)
        if ret != 0:
            raise RuntimeError(f"Error in {self.entrypoint}: {ret}")
        return ret






from dataclasses import dataclass
from math import prod
import torch as th
from vidrial.jit.static import Shape, Layout, Int, Tuple
from vidrial.jit.static.util import torch_dtype_to_c
from vidrial.jit.static.util import layout_from_shape_stride
from vidrial.jit.jit import jit, render
from vidrial.jit.timingcache import ConfigTimingCache
from vidrial.jit.binding import make_binding
from vidrial.kernels.sympow.dimensions import op_output_shape, sympow_dim, problem_dimensions
from vidrial.kernels.sympow.binding import configurations as fwd_configurations

# ------------------- Source Code -------------------

@dataclass
class SourceCode:
    T: str
    XSlabShape: Shape # [d,[b1,...]]
    p: int
    d_tile: int
    duplicate_correction: bool
    GXSlab: Layout
    GZSlab: Layout
    b_tile: int
    ZFrgShape: Shape
    SZTileLayout: Layout
    smem_acc: Tuple
    @property
    def template(self):
        return """
#include <cuda.h>
#include <cuda_runtime.h>
#include <iostream>
#include "cute/tensor.hpp"
#include "cutlass/cutlass.h"
#include "kernels/sympow_bwd/sympow_bwd.cuh"
#include "kernels/sympow_bwd/kernel.cuh"

using namespace cute;
using namespace vidrial;

extern "C" int launch(void* __raw_X, void* __raw_Zgrad, void* __raw_Xgrad) {
    using T = {T};
    auto X = reinterpret_cast<T*>(__raw_X);
    auto Zgrad = reinterpret_cast<T*>(__raw_Zgrad);
    auto Xgrad = reinterpret_cast<T*>(__raw_Xgrad);
    using GXSlab = {GXSlab};
    using GZSlab = {GZSlab};
    constexpr int p = {p};
    constexpr int d_tile = {d_tile};
    constexpr int b_tile = {b_tile};
    using ZFrgShape = {ZFrgShape};
    constexpr bool duplicate_correction = {duplicate_correction};
    using XSlabShape = {XSlabShape};
    using ZSlabShape = decltype(sympow_shape<p, d_tile>(XSlabShape{}));
    using XTileShape = Shape<Int<d_tile>, Int<b_tile>>; 
    using ZTileShape = decltype(tpow_shape<p>(XTileShape{}));
    using SZTileLayout = {SZTileLayout};
    using ZFrgThr = decltype(zipped_divide(Layout<ZTileShape>{}, ZFrgShape{}));
    auto cfg = make_sympow_bwd_cfg<T, p, {smem_acc}>(ZFrgShape{}, XTileShape{}, GZSlab{}, GXSlab{}, SZTileLayout{});
    return launch_tiled_sympow_bwd_kernel<duplicate_correction>(cfg, X, Zgrad, Xgrad);
}
"""
    def __str__(self):
        return render(self.template, self.__dict__)

# ------------------- Binding -------------------
@dataclass
class BindingCfg:
    d: int
    bs: tuple[int, ...]
    X_shape: tuple[int, ...]
    Z_shape: tuple[int, ...]
    X_stride: tuple[int, ...]
    Z_stride: tuple[int, ...]
    power: int
    d_tile: int
    duplicate_correction: bool
    # Performance parameters (don't affect the output)
    b_tile: int
    ZFrgShape: Shape
    SZTileLayout: Layout
    dtype: th.dtype
    smem_acc: tuple[bool, ...]
    @classmethod
    def from_args(cls, X, Zgrad, Xgrad, power, d_tile, duplicate_correction, b_tile, ZFrgShape, SZTileLayout, smem_acc):
        assert X.device.type == Zgrad.device.type == Xgrad.device.type == 'cuda', f"Invalid {X.device}, {Zgrad.device}, {Xgrad.device}."
        assert X.dtype == Zgrad.dtype == Xgrad.dtype
        dtype = X.dtype
        assert X.shape == Xgrad.shape, "X and Xgrad must have the same shape"
        assert X.stride() == Xgrad.stride(), "X and Xgrad must have the same stride"
        bs, d, D = problem_dimensions(Zgrad.shape, X.shape, power, d_tile)
        assert Zgrad.shape == op_output_shape(X.shape, power, d_tile)
        return BindingCfg(d, bs, X.shape, Zgrad.shape, X.stride(), Zgrad.stride(),
                          power, d_tile, duplicate_correction, b_tile, ZFrgShape, SZTileLayout, dtype, smem_acc)
    @property
    def source(self):
        """ Conversion between the different variable name and shape conventions
                D = sympow_dim(d, power, d_tile)
                    Python                     c++
                X_shape=[P,D]      ->      GXSlabShape=[D,P]
                Z_shape=[P,tile_num, d_tile, ...]  ->  GZSlabShape=[[d_tile,...],tile_num],P]
        """
        p = self.power
        XSlabShape = Shape(Int(self.d), Shape(*[Int(b) for b in self.bs]))
        batch_num = len(self.bs)
        batch_dims = tuple(range(batch_num))
        X_py2cuda = (batch_num, batch_dims)
        Z_py2cuda = ((tuple(range(p+batch_num, batch_num, -1)), batch_num), batch_dims)
        GXSlab = layout_from_shape_stride(self.X_shape, self.X_stride, X_py2cuda)
        GZSlab = layout_from_shape_stride(self.Z_shape, self.Z_stride, Z_py2cuda)
        smem_acc = Tuple(*[Int(1 if smem_acc_i else 0) for smem_acc_i in self.smem_acc])
        return SourceCode(
            T=torch_dtype_to_c(self.dtype),
            XSlabShape=XSlabShape,
            p=self.power,
            d_tile=self.d_tile,
            duplicate_correction=self.duplicate_correction,
            GXSlab=GXSlab,
            GZSlab=GZSlab,
            b_tile=self.b_tile,
            ZFrgShape=self.ZFrgShape,
            SZTileLayout=self.SZTileLayout,
            smem_acc=smem_acc
        )


# ---------------------- Autotune --------------------------------
def configurations(args: dict) -> list[dict]:
    X, Zgrad, power, d_tile = args['X'], args['Zgrad'], args['power'], args['d_tile']
    configs = fwd_configurations(dict(Z=Zgrad, X=X, power=power, d_tile=d_tile))
    for cfg in configs:
        cfg['smem_acc'] = [True] * power # by default all dimensions are accumulated in smem
        cfg['smem_acc'][0] = False # except for the last one, which accumulates in registers
    return configs

def hash_fn(args: dict) -> str:
    Zgrad, X, power, d_tile = args['Zgrad'], args['X'], args['power'], args['d_tile']
    bs,d,D = problem_dimensions(Zgrad.shape, X.shape, power, d_tile)
    return f"Zgrad.shape={Zgrad.shape}-X.shape={X.shape}-power={power}-d_tile={d_tile}-D={D}-d={d}"

@make_binding(cache=ConfigTimingCache('sympow_bwd', hash_fn),
          sweep=configurations)
def binding(X, Zgrad, Xgrad, power, d_tile, duplicate_correction, b_tile, ZFrgShape, SZTileLayout, smem_acc):
    binding_cfg = BindingCfg.from_args(X, Zgrad, Xgrad, power, d_tile, duplicate_correction, b_tile, ZFrgShape, SZTileLayout, smem_acc)
    jit(name = "sympow_bwd",
        code = str(binding_cfg.source),
    )(X, Zgrad, Xgrad)






"""Data types for C++ type system representation.

This module provides Python classes that represent C++ types and can generate
their C++ string representations. Used for JIT compilation and type system
manipulation.
"""

from .types import (
    # Enums
    CudaDType,
    
    # Core types
    Int,
    Shape,
    Stride,
    Layout,
    
    # Traits
    Trait,
    UniversalCopy,
    CopyAtom,

    # Common trait instances
    UniversalCopyFloat,
    UniversalCopyUint8,

    # Layouts
    compact_major,
)
from .util import layout_from_tensor
from .algorithms import merge, get
from .tuples import flatten, Tuple
from .int import _1, _2, _3, _4, _5, _6, _7, _8, _16, _32, _64, _128, _256

__all__ = [
    'CudaDType',
    'Int',
    'Shape',
    'Stride',
    'Layout',
    'Tuple',
    'Trait',
    'UniversalCopy',
    'CopyAtom',
    'flatten',
    'merge',
    'get',
    'compact_major',
    'layout_from_tensor',
    'UniversalCopyFloat',
    'UniversalCopyUint8',
    '_1',
    '_2',
    '_3',
    '_4',
    '_5',
    '_6',
    '_7',
    '_8',
    '_16',
    '_32',
    '_64',
    '_128',
    '_256',
]



from dataclasses import dataclass
from functools import lru_cache
import torch as th
from math import prod
from vidrial.jit.static.types import Shape, Layout, Int
from vidrial.jit.static.util import torch_dtype_to_c
from vidrial.jit.static.util import layout_from_shape_stride, nested_constructor
from vidrial.jit.jit import jit, render
from vidrial.jit.timingcache import ConfigTimingCache
from vidrial.jit.binding import make_binding
from vidrial.kernels.mma_configurator import dtype_to_bytes


# ------------------- Source Code -------------------
@dataclass
class SourceCode:
    T: str
    fn_str: str # c++ for the reduction function
    GXSlab: Layout
    GxSlab: Layout
    thread_num: int
    XTileShape: Shape
    @property
    def template(self):
        return """
# include <cuda.h>
# include <cuda_runtime.h>
#include <cute/tensor.hpp>
#include "kernels/reduce/kernel.cuh"
#include "kernels/reduce/reduce_utils.cuh"
using namespace cute;
using namespace vidrial;
using T = {T};
struct Fn {
    template<typename Ta, typename Tb> CUTE_HOST_DEVICE
    auto operator()(Ta a, Tb b) {
        {fn_str}
    }
};
extern "C" void launch(void* _X, void* _x) {
    auto X = reinterpret_cast<T*>(_X);
    auto x = reinterpret_cast<T*>(_x);
    auto cfg = ReduceKernelCfg<T, {thread_num}, {XTileShape}, {GXSlab}, {GxSlab}>{};
    launch_reduce_kernel(cfg, X, x, Fn{});
}
"""
    def __str__(self):
        return render(self.template, self.__dict__)

# ------------------- Binding -------------------
@dataclass
class BindingCfg:
    fn_str: str
    reduce_dims: tuple[int, ...]
    X_shape: tuple[int, ...]
    x_shape: tuple[int, ...] # Same dims as X_shape but with the reduce_dim removed
    X_stride: tuple[int, ...]
    x_stride: tuple[int, ...]
    X_tile_shape: tuple[int, ...]
    thread_num: int
    dtype: th.dtype
    @classmethod
    def from_args(cls, X, x, fn, reduce_dims, X_tile_shape, thread_num):
        assert X.device.type == x.device.type == 'cuda', f"Invalid {X.device}, {x.device}."
        assert X.dtype == x.dtype
        dtype = X.dtype
        return cls(fn, reduce_dims, X.shape, x.shape, X.stride(), x.stride(), X_tile_shape, thread_num, dtype)
    @property
    def source(self):
        n = len(self.X_shape)
        reduce_dims = [i for i in range(n) if i in self.reduce_dims]
        batch_dims = [i for i in range(n) if i not in self.reduce_dims]
        X_dims = (reduce_dims, batch_dims)
        GXSlab = layout_from_shape_stride(self.X_shape, self.X_stride, X_dims)
        GxSlab = layout_from_shape_stride(self.x_shape, self.x_stride)
        XTileShape = nested_constructor(X_dims)(self.X_tile_shape, Shape)
        sc = SourceCode(
            T=torch_dtype_to_c(self.dtype),
            fn_str=self.fn_str,
            GXSlab=GXSlab,
            GxSlab=GxSlab,
            thread_num=self.thread_num,
            XTileShape=XTileShape,
        )
        return sc

def configurations(binding_args: dict) -> list[dict]:
    """ Given any args to binding, 
    Currently returns a single reasonable configuration.
    TODO: Return a better configuration sweep
    """
    rest_args = {}
    if 'X_tile_shape' not in binding_args:
        X_tile_shape = [min(x, 16) for x in binding_args['X'].shape]
        rest_args['X_tile_shape'] = X_tile_shape
    if 'thread_num' not in binding_args:
        rest_args['thread_num'] = 32
    return [rest_args]

def sweep_key(binding_args: dict) -> str:
    X, x, reduce_dims = binding_args['X'], binding_args['x'], binding_args['reduce_dims']
    msg = f"X:shape={X.shape},stride={X.stride()}-"
    msg += f"x:shape={x.shape},stride={x.stride()}-"
    msg += f"reduce_dims={reduce_dims}"
    return msg

@make_binding(cache=ConfigTimingCache('reduce', sweep_key),
          sweep=configurations)
def binding(X, x, fn, reduce_dims, X_tile_shape, thread_num):
    binding_cfg = BindingCfg.from_args(X, x, fn, reduce_dims, X_tile_shape, thread_num)
    jit(name = "reduce",
        code = str(binding_cfg.source),
    )(X, x)



import os
from typing import Any
import shutil
import logging
import subprocess
from tqdm import tqdm
import yaml

logger = logging.getLogger(__name__)

FUNCTIONS = []
CONFIG_CACHES = []


def register_runtime(name: str, path: str):
    """
    Register a runtime for a given function. This is used for packaging built
    runtimes for release.

    Args:
        name: The name of the function.
        path: The path to the runtime.
    """
    if os.environ.get('JIT_PACKAGE_MODE', '0') == '1':  
        FUNCTIONS.append(dict(name=name, path=path))


def register_config_cache(cache: Any):
    """
    Register a config cache for a given function. This is used for packaging
    config caches for release.

    Args:
        name: The name of the function.
        cache: The config cache.
    """
    if os.environ.get('JIT_PACKAGE_MODE', '0') == '1':
        CONFIG_CACHES.append(cache)


def get_git_status():
    """ Get the current git status.

    Returns:
        A dictionary containing the git status.
    """
    commands = {
        "root": ["git", "rev-parse", "--show-toplevel"],
        "commit": ["git", "rev-parse", "HEAD"],
        "status": ["git", "status", "--porcelain"],
    }
    info = {}
    for key, cmd in commands.items():
        try:
            info[key] = subprocess.check_output(cmd, cwd=os.getcwd(), text=True).strip()
        except subprocess.CalledProcessError:
            info[key] = "unknown"
    info['status'] = "clean" if info['status'] == "" else "dirty"
    return info


def package_release(version: str, root: str):
    """
    Package the current state of JIT functions for release.

    Args:
        version: The version string to use for the package
        root: The root directory to store the package
    """
    if os.environ.get('JIT_PACKAGE_MODE', '0') == '1':
        os.makedirs(root, exist_ok=True)

    for cache in tqdm(CONFIG_CACHES, desc="Packaging config caches"):
        cache.store(os.path.join(root, f"{cache.fn}.yaml"))
        logger.info(f"Packaged config cache for {cache.fn} to {os.path.join(root, f'{cache.fn}.yaml')}")

    for function in tqdm(FUNCTIONS, desc="Packaging runtimes"):
        shutil.copytree(function["path"], os.path.join(root, os.path.basename(function["path"])), dirs_exist_ok=True)
        logger.info(f"Packaged runtime for {function['name']} to {os.path.join(root, os.path.basename(function['path']))}")

    status = get_git_status()
    status["version"] = version
    with open(os.path.join(root, "version.yaml"), "w") as f:
        yaml.dump(status, f)

    logger.info(f"Kernels packaged to {root}")



from vidrial.kernels.sympow.op import op, op_reference

def interface(X, power, d_tile=1, dim=-1, duplicate_correction=True):
    """ Computes the tiled symmetric power of X.
    Dimensions:
        d the feature dimension being powered.
        b the batch dimensions
        D the expanded dimension. Given by sympow_dim(d, power, d_tile)
    Args:
        X: Input tensor of shape `[d0, d1, ...]`.
        power: Power of the symmetric power.
        d_tile: Tile size for the operation.
        duplicate_correction: Whether to correct for duplicate elements.
    Returns:
        Z: Same shape as X except that the dimension at position `dim` is expanded to D
           for example, if dim=0 Z is [D, d1, d2, ...]
    """
    n = len(X.shape)
    if dim >= 0: dim -= n
    assert dim < 0
    X = X.transpose(-1, dim) # Place the expand dim at the end (convention used in the op)
    Z = op(X, power, d_tile, duplicate_correction)
    Z = Z.reshape(X.shape[:-1] + (-1,)) # Flatten tile_count, d_tile, d_tile, ...
    return Z.transpose(-1, dim)

def interface_reference(X, power, d_tile=1, dim=-1, duplicate_correction=True):
    n = len(X.shape)
    if dim >= 0: dim -= n
    assert dim < 0
    X = X.transpose(-1, dim) # Place the expand dim at the end (convention used in the op)
    Z = op_reference(X, power, d_tile, duplicate_correction)
    Z = Z.reshape(X.shape[:-1] + (-1,)) # Flatten tile_count, d_tile, d_tile, ...
    return Z.transpose(-1, dim)


from typing import Dict, Any
from dataclasses import dataclass
from math import prod, log2
import torch as th
import logging
from vidrial.jit.static.types import Shape, Layout, Int
from vidrial.jit.static.util import layout_from_shape_stride, torch_dtype_to_c
from vidrial.py_utils.test_utils import diff
from vidrial.kernels.mma_configurator import advanced_configurator
from vidrial.jit.timingcache import ConfigTimingCache
from vidrial.jit.binding import make_binding
from vidrial.jit.jit import jit, render
from vidrial.kernels.sympow_mma.dimensions import problem_shape

logger = logging.getLogger(__name__)


# ------------------- Source Code -------------------

@dataclass
class SourceCode:
    T: str
    MNKPSlabShape: Shape
    MNKTileShape: Shape
    d: int
    d_tile: int
    pow: int
    expand_K: bool
    GaSlab: Layout
    GBSlab: Layout
    GCSlab: Layout
    duplicate_correction: bool
    smempipe: int
    regpipe: int
    use_ldsm: bool
    swizzle: int
    Atom: str
    MNKAtomPlacement: Shape
    scale_A: bool
    @property
    def template(self):
        return """
#include <cuda.h>
#include <cuda_runtime.h>
#include <iostream>
#include "cute/tensor.hpp"
#include "kernels/sympow_mma/kernel.cuh"

using namespace cute;
using namespace vidrial;

extern "C" int launch(void* __raw_A, void* __raw_B, void* __raw_C, float sA) {
    using T = {T};
    using MNKPSlabShape = decltype(static_tree_cast<int64_t>({MNKPSlabShape}{}));
    using MNKTileShape = decltype(static_tree_cast<int64_t>({MNKTileShape}{}));
    constexpr int d = {d}, d_tile = {d_tile};
    constexpr int pow = {pow};
    constexpr bool expand_K = {expand_K};
    using GaSlab = decltype(static_tree_cast<int64_t>({GaSlab}{}));
    using GBSlab = decltype(static_tree_cast<int64_t>({GBSlab}{}));
    using GCSlab = decltype(static_tree_cast<int64_t>({GCSlab}{}));
    constexpr bool duplicate_correction = {duplicate_correction};
    using PerfCfg = PerfCfg<{smempipe}, {regpipe}, {use_ldsm}, {swizzle}>;
    using Atom = MMA_Atom<{Atom}>;
    using MNKAtomPlacement = decltype(static_tree_cast<int64_t>({MNKAtomPlacement}{}));
    using Cfg = SympowMmaKernelCfg<T, pow, Atom, MNKAtomPlacement, MNKPSlabShape, MNKTileShape, d, d_tile, expand_K, GaSlab, GBSlab, GCSlab, PerfCfg>;
    auto A = reinterpret_cast<T*>(__raw_A);
    auto B = reinterpret_cast<T*>(__raw_B);
    auto C = reinterpret_cast<T*>(__raw_C);
    return launch_sympow_mma_kernel<duplicate_correction, {scale_A}>(Cfg{}, A, B, C, sA);
}"""
    def __str__(self):
        return render(self.template, self.__dict__)


# ------------------- Binding -------------------

@dataclass
class BindingCfg:
    batches: tuple[int, ...]
    M: int
    N: int
    K: int
    d: int
    d_tile: int
    scale_A: bool
    pow: int
    expand_dim: int 
    A_shape: tuple[int, ...]
    B_shape: tuple[int, ...]
    C_shape: tuple[int, ...]
    A_stride: tuple[int, ...]
    B_stride: tuple[int, ...]
    C_stride: tuple[int, ...]
    duplicate_correction: bool
    smempipe: int
    regpipe: int
    use_ldsm: bool
    swizzle: int
    Atom: str
    MNKTileShape: Shape 
    MNKAtomPlacement: Shape
    dtype: th.dtype
    def is_valid(self):
        assert self.M % self.MNKTileShape[0].value == 0 # type: ignore
        assert self.N % self.MNKTileShape[1].value == 0 # type: ignore
        assert self.K % self.MNKTileShape[2].value == 0 # type: ignore

    @classmethod
    def from_args(cls, A: th.Tensor, B: th.Tensor, C: th.Tensor, expand_dim: int, power: int, d_tile: int, scale_A: float, duplicate_correction: bool,
                  smempipe: int, regpipe: int, use_ldsm: bool, swizzle: int, Atom: str, MNKTileShape: Shape, MNKAtomPlacement: Shape) -> 'BindingCfg':
        assert A.device.type == B.device.type == C.device.type == 'cuda', f"Invalid {A.device}, {B.device}, {C.device}."
        assert A.dtype == B.dtype == C.dtype, f"Invalid {A.dtype}, {B.dtype}, {C.dtype}. Kernel currently assumes all input tensors have the same dtype"
        dtype = A.dtype
        batches, M, N, K, d, D = problem_shape(A.shape, B.shape, C.shape, expand_dim, power, d_tile)
        self = cls(batches=batches, M=M, N=N, K=K, d=d, d_tile=d_tile, pow=power, expand_dim=expand_dim, scale_A=(scale_A != 1.0), duplicate_correction=duplicate_correction, 
                   A_shape=A.shape, B_shape=B.shape, C_shape=C.shape,
                   A_stride=A.stride(), B_stride=B.stride(), C_stride=C.stride(),
                   smempipe=smempipe, regpipe=regpipe, use_ldsm=use_ldsm, swizzle=swizzle,
                   Atom=Atom, MNKTileShape=MNKTileShape, MNKAtomPlacement=MNKAtomPlacement, dtype=dtype)
        self.is_valid()
        logger.debug("BindingCfg: %s", self)
        return self

    @property
    def source(self):
        """ Conversion between the different variable name and shape conventions
                     Python                       c++
                A                    ->      a
                eA                   ->      A
                A_shape=[P,M,d]      ->      aSlabShape=[M,d,P]    if expand_K
                A_shape=[P,d,K]      ->      aSlabShape=[d,K,P]    if not expand_K
                B_shape=[P,K,N]      ->      BSlabShape=[N,K,P]
                C_shape=[P,M,N]      ->      CSlabShape=[M,N,P]
        """
        batch_inds = tuple(range(len(self.batches)))
        PShape = Shape(*[Int(b) for b in self.batches])
        source = SourceCode(
            T=torch_dtype_to_c(self.dtype),
            MNKPSlabShape=Shape(Int(self.M), Int(self.N), Int(self.K), PShape),
            d=self.d,
            d_tile=self.d_tile,
            pow=self.pow,
            expand_K=True if self.expand_dim%3 == 2 else False,
            GaSlab=layout_from_shape_stride(self.A_shape, self.A_stride, (-2, -1, batch_inds)),
            GBSlab=layout_from_shape_stride(self.B_shape, self.B_stride, (-1, -2, batch_inds)),
            GCSlab=layout_from_shape_stride(self.C_shape, self.C_stride, (-2, -1, batch_inds)),
            duplicate_correction=self.duplicate_correction,
            smempipe=self.smempipe,
            regpipe=self.regpipe,
            use_ldsm=self.use_ldsm,
            swizzle=self.swizzle,
            Atom=self.Atom,
            MNKTileShape=self.MNKTileShape,
            scale_A=self.scale_A,
            MNKAtomPlacement=self.MNKAtomPlacement)
        return source

# ---------------------- Autotune --------------------------------

def make_configurator(smempipe: tuple[int, int], regpipe: tuple[int, int], use_ldsm: bool, swizzle: int):
    def configurator(args: dict) -> list[Dict[str, Any]]:
        A, B, C, expand_dim, power, d_tile, duplicate_correction = args['A'], args['B'], args['C'], args['expand_dim'], args['power'], args['d_tile'], args['duplicate_correction']
        batches, M, N, K, d, D = problem_shape(A.shape, B.shape, C.shape, expand_dim, power, d_tile)
        D_tile = d_tile ** power
        if expand_dim % 3 == 2: # expand K
            M_tile, N_tile, K_tile = None, None, D_tile 
        else: # expand M
            M_tile, N_tile, K_tile = D_tile, None, None
        configs = advanced_configurator((M, N, K, batches), A.dtype, th.float32, M_tile, N_tile, K_tile, smempipe=smempipe, regpipe=regpipe, use_ldsm=use_ldsm, swizzle=swizzle)
        return configs
    return configurator

def hash_fn(args: dict) -> str:
    P, M, N, K, d, D = problem_shape(args['A'].shape, args['B'].shape, args['C'].shape, args['expand_dim'], args['power'], args['d_tile'])
    A, B, C = args['A'], args['B'], args['C']
    key = f"M_{int(log2(M))}-N_{int(log2(N))}-K_{int(log2(K))}-"
    key += f"{A.dtype}-{args['expand_dim'] % A.ndim}-{args['power']}-{args['d_tile']}-"
    find_major_dim = lambda x: [i for i, s in enumerate(x.stride()) if s == 1][0]
    key += f"{find_major_dim(A)}-{find_major_dim(B)}-{find_major_dim(C)}"
    return key


@make_binding(cache=ConfigTimingCache('sympow_mma', hash_fn), 
          sweep=make_configurator(smempipe=(1,2), regpipe=(0,1), use_ldsm=True, swizzle=1))
def binding(A, B, C, expand_dim, power, d_tile, scale_A, duplicate_correction, smempipe, regpipe, use_ldsm, swizzle, Atom, MNKTileShape, MNKAtomPlacement):
    binding_cfg = BindingCfg.from_args(A, B, C, expand_dim, power, d_tile, scale_A, duplicate_correction, smempipe, regpipe, use_ldsm, swizzle, Atom, MNKTileShape, MNKAtomPlacement)
    jit(name = "sympow_mma",
        code = str(binding_cfg.source)
    )(A, B, C, scale_A)



import torch as th
from math import sqrt
from vidrial.kernels.sympow.dimensions import op_output_shape as fwd_op_output_shape, problem_dimensions, SympowCoords
from vidrial.kernels.sympow_bwd.binding import binding
from vidrial.py_utils.common import tprod_bwd


# ---------------------- Op --------------------------------

@th.library.custom_op('vidrial::sympow_bwd', mutates_args=(), device_types='cuda')
def op(X: th.Tensor, Zgrad: th.Tensor, power: int, d_tile: int, duplicate_correction: bool = True) -> th.Tensor:
    """ Computes gradient wrt X of Z = sympow(X)
    Inputs:
        X: [b0, b1, ..., d] - Batch dimensions, feature dimension
        Zgrad: [b0, b1, ..., tn, dt, dt, ...] where dt = d_tile and tn is the tile number
        power: int - Power of the symmetric power
        d_tile: int - Tile size
        duplicate_correction: bool - Whether to use duplicate correction
    Output shape:
        Xgrad: [b0, b1, ..., d] - Gradient with respect to input X
    """
    Xgrad = th.empty_strided(X.size(), X.stride(), dtype=X.dtype, device=X.device)
    binding(X, Zgrad, Xgrad, power, d_tile, duplicate_correction)
    return Xgrad

# Fake implementation for tracing and testing
@op.register_fake
def op_fake(X: th.Tensor, Zgrad: th.Tensor, power: int, d_tile: int, duplicate_correction: bool = True):
    return th.empty_like(X)


# ---------------------- Reference Op  --------------------------------

def op_reference(X: th.Tensor, Zgrad: th.Tensor, power: int, d_tile: int, duplicate_correction: bool = True) -> th.Tensor:
    Xgrad = th.zeros_like(X)
    bs, d, D = problem_dimensions(Zgrad.shape, X.shape, power, d_tile)
    batch_slice = (slice(None),) * len(bs)
    X_tiled = X.reshape(bs + (-1, d_tile))
    Xgrad_tiled = Xgrad.reshape(bs + (-1, d_tile))
    for idx, seq, dup in SympowCoords(d // d_tile, power):
        Zgrad_tile = Zgrad[batch_slice + (idx,)].clone()
        if duplicate_correction: Zgrad_tile[:] *= sqrt(dup)
        X_tiles = [X_tiled[batch_slice + (seq[i],)] for i in range(power)]
        Xgrad_tiles = tprod_bwd(X_tiles, Zgrad_tile, power)
        for i in range(power):
            Xgrad_tile = Xgrad_tiled[batch_slice + (seq[-i-1],)]
            Xgrad_tile += Xgrad_tiles[i]
    return Xgrad




import pytest
import torch as th
from vidrial.kernels.sympow.dimensions import op_output_shape as fwd_op_output_shape
from vidrial.kernels.sympow.op import op_reference as fwd_op_reference
from vidrial.kernels.sympow_bwd.binding import binding
from vidrial.kernels.sympow_bwd.op import op, op_reference
from vidrial.py_utils.test_utils import  PERMUTATIONS_OF_STRIDES_UP_TO_4D, create_permuted_strides_layout
from vidrial.jit.static.types import Shape, Layout, Int
from vidrial.py_utils.test_utils import diff

XSHAPES_AND_POWERS_AND_TILES = [
    ((1, 4), 2, 1),
    ((1, 4), 3, 1),
    ((1, 4), 3, 2),
    ((1, 8), 3, 2),
    ((1, 8), 3, 4),
    ((1, 8), 4, 4),
    ((4, 4), 3, 1),
    ((4, 4), 3, 2),
    ((4, 8), 3, 2),
    ((4, 8), 3, 4),
    ((1, 4), 2, 4),
    ((1, 8), 2, 4),
    ((1, 4), 3, 2),
    ((16, 32), 2, 4),
    ((2, 16, 32), 2, 4),
    ((16, 32), 3, 4),
    ((16, 32), 4, 4),
    ((128, 64), 2, 4),
]

def test_call_binding():
    X = th.randn(8, 64, device='cuda')
    Zgrad = th.randn(fwd_op_output_shape(X.shape, 2, 8), device='cuda')
    Xgrad = th.empty(X.shape, device='cuda')
    binding(X, Zgrad, Xgrad, 2, 8, True, 1,
            ZFrgShape=Shape(Shape(Int(8), Int(1)), Int(1)),
            SZTileLayout=Layout(Shape(Shape(Int(8), Int(8)), Int(1))),
            smem_acc=[True, True])

@pytest.mark.parametrize("X_shape, power, d_tile", XSHAPES_AND_POWERS_AND_TILES)
@pytest.mark.parametrize("duplicate_correction", [True, False])
def test_binding_autotuned(X_shape, power, d_tile, duplicate_correction):
    Z_shape = fwd_op_output_shape(X_shape, power, d_tile)
    X = th.randn(X_shape, device="cuda")
    Zgrad = th.randn(Z_shape, device="cuda" )
    Xgrad = th.zeros(X_shape, device="cuda")
    binding(X, Zgrad, Xgrad, power, d_tile, duplicate_correction)

@pytest.mark.parametrize("X_shape, power, d_tile", XSHAPES_AND_POWERS_AND_TILES)
@pytest.mark.parametrize("duplicate_correction", [True])
def test_op_vs_ref(X_shape, power, d_tile, duplicate_correction):
    Z_shape = fwd_op_output_shape(X_shape, power, d_tile)
    X = th.randn(X_shape, device="cuda")
    Zgrad = th.randn(Z_shape, device="cuda" )
    Xgrad = op(X, Zgrad, power, d_tile, duplicate_correction)
    Xgrad_ref = op_reference(X, Zgrad, power, d_tile, duplicate_correction)
    diff(Xgrad, Xgrad_ref)

@pytest.mark.parametrize("X_shape, power, d_tile", XSHAPES_AND_POWERS_AND_TILES)
@pytest.mark.parametrize("perm", PERMUTATIONS_OF_STRIDES_UP_TO_4D)
def test_op_generic_strides(X_shape, power, d_tile, perm):
    X = th.randn(X_shape, device="cuda")
    Zgrad = th.randn(fwd_op_output_shape(X_shape, power, d_tile), device="cuda")
    X = create_permuted_strides_layout(X, perm)
    Zgrad = create_permuted_strides_layout(Zgrad, perm)
    # Compare op with op_reference
    Xgrad = op(X, Zgrad, power, d_tile, True)
    Xgrad_ref = op_reference(X, Zgrad, power, d_tile, True)
    diff(Xgrad, Xgrad_ref)

@pytest.mark.parametrize("X_shape, power, d_tile", XSHAPES_AND_POWERS_AND_TILES)
@pytest.mark.parametrize("duplicate_correction", [True, False])
def test_op_vs_fwd_ref_bwd(X_shape, power, d_tile, duplicate_correction):
    """Compares op_reference against the autograd of fwd_op_reference """
    Z_shape = fwd_op_output_shape(X_shape, power, d_tile)
    X = th.randn(X_shape, device="cuda", requires_grad=True)
    Zgrad = th.randn(Z_shape, device="cuda" )
    Z = fwd_op_reference(X, power, d_tile, duplicate_correction)
    Z.backward(Zgrad)
    Xgrad_ref = X.grad.clone()
    Xgrad = op(X, Zgrad, power, d_tile, duplicate_correction)
    diff(Xgrad, Xgrad_ref)


import torch as th
import pytest
from vidrial.kernels.sympow.binding import binding, binding_autotuned
from vidrial.kernels.sympow.dimensions import sympow_dim, op_output_shape, interface_output_shape
from vidrial.jit.static.types import Shape, Layout, Int
from vidrial.kernels.sympow.op import op, op_reference
from vidrial.py_utils.test_utils import diff
from vidrial.kernels.sympow.interface import interface
from vidrial.py_utils.common import tpow_shape, tpow
from vidrial.py_utils.test_utils import PERMUTATIONS_OF_STRIDES_UP_TO_4D, create_permuted_strides_layout

# Basic binding, op, and interface tests
def test_call_binding():
    b, d = 128, 64
    power, d_tile = 2, 8
    X = th.empty((b, d), device='cuda')
    Z = th.empty(op_output_shape(X.shape, power, d_tile), device='cuda')
    binding(Z, X, power, d_tile, duplicate_correction=True, b_tile=4,
            ZFrgShape=Shape(Shape(*[Int(2)]*power), Int(1)),
            SZTileLayout=Layout(Shape(Shape(*[Int(d_tile)]*power), Int(4))))
def test_sympow_op():
    b, d = 128, 64
    power, d_tile = 2, 8
    X = th.empty((b, d), device='cuda')
    Z = th.empty(op_output_shape(X.shape, power, d_tile), device='cuda')
    Z = op(X, power, d_tile, duplicate_correction=True)
    Z_ref = op_reference(X, power, d_tile, duplicate_correction=True)
    diff(Z, Z_ref, rtol=4e-2, atol=1e-2)
def test_sympow_interface():
    X = th.empty((8, 16), device='cuda')
    power, d_tile = 2, 8
    Z = interface(X, power, d_tile, dim=0)
    Z = interface(X, power, d_tile, dim=1)

# ----------- Shape Tests -----------
# TODO: move tpow tests somewhere else
@pytest.mark.parametrize("x_shape,expected,p", [
    ([2], [2,2], 2),
    ([3], [3,3], 2),
    ([5], [5,5], 2),
    ([9], [9,9], 2),
    ([2,4], [2,4,4], 2), # batch dims
    ([2,2], [2,2,2], 2),
    ([1,7,5], [1,7,5,5], 2),
    ([2,3,4,8], [2,3,4,8,8], 2),
    ([5], [5,5,5], 3),
    ([9], [9,9,9], 3),
])
def test_tpow_shape(x_shape, expected, p):
    assert tpow_shape(x_shape, p) == th.Size(expected)
    assert tpow_shape(th.Size(x_shape), p) == th.Size(expected)

@pytest.mark.parametrize("x_shape,expected,p", [
    ([2], [3], 2),
    ([3], [6], 2),
    ([4], [10], 2),
    ([5], [15], 2),
    ([7], [28], 2),
    ([8], [36], 2),
    ([9], [45], 2),
    ([10], [55], 2),
    ([2,4], [2,10], 2), # batch dims
    ([1,7,5], [1,7,15], 2),
    ([2,3,4,8], [2,3,4,36], 2),
])
def test_interface_output_shape(x_shape, expected, p):
    assert interface_output_shape(x_shape, p) == th.Size(expected)
    assert interface_output_shape(th.Size(x_shape), p) == th.Size(expected)


# # ----------- Kernel Tests -----------
SHAPES_AND_POWERS_AND_TILES = [
    ([2], 2, 1),
    ([4], 2, 2),
    ([4], 3, 2),
    ([8], 3, 4),
    ([3,8], 3, 4),
    ([3,8], 2, 4),
    ([3,2], 4, 1),
    ([2,3,8], 2, 1),
    ([2,3,16], 3, 8),
]
LARGE_SHAPES_AND_POWERS_AND_TILES = [
    # ([32], 2, 4),
    ([2, 32], 2, 4),
    # ([32], 3, 4),
    # ([64], 2, 8),
    ([8,16], 2, 8),
    # ([128,16], 3, 4),
    # ([128,32], 4, 2),
    ([128,32], 3, 1),
    ([128,4,32], 4, 2),
    # ([64,32,16], 2, 1), # fails with > 1 batch dimension due to some cpp error
    # ([64,32,8], 3, 8),  # fails with > 1 batch dimension due to some cpp error
]


@pytest.mark.parametrize("X_shape,p,_", SHAPES_AND_POWERS_AND_TILES)
def test_tpow(X_shape, p, _):
    x = th.empty(X_shape)
    Y = tpow(x, p)
    assert Y.shape == tpow_shape(X_shape, p)

def _test_op(X_shape, p, d_tile, duplicate_correction):
    X = th.randn(X_shape, device='cuda')
    Z = op(X, p, d_tile, duplicate_correction)
    assert Z.shape == op_output_shape(X_shape, p, d_tile)
    Z_ref = op_reference(X, p, d_tile, duplicate_correction)
    diff(Z, Z_ref, rtol=4e-2, atol=1e-2)
@pytest.mark.parametrize("X_shape,p,d_tile", SHAPES_AND_POWERS_AND_TILES)
@pytest.mark.parametrize("duplicate_correction", [True, False])
def test_op_small(X_shape, p, d_tile, duplicate_correction):
    _test_op(X_shape, p, d_tile, duplicate_correction)
@pytest.mark.parametrize("X_shape,p,d_tile", LARGE_SHAPES_AND_POWERS_AND_TILES)
def test_op_large(X_shape, p, d_tile):
    _test_op(X_shape, p, d_tile, True)

@pytest.mark.parametrize("X_shape,p,d_tile", SHAPES_AND_POWERS_AND_TILES)
def test_sympow_tpow_norm_equivalence(X_shape, p, d_tile):
    X = th.randn(X_shape, device='cuda')
    Y = tpow(X, p).flatten(start_dim=-p)
    Z = interface(X, p, d_tile, -1)
    diff(th.linalg.norm(Y, dim=-1), th.linalg.norm(Z, dim=-1), rtol=4e-2, atol=1e-2)
 

def _test_interface(X_shape, power, d_tile, dim):
    """Test SympowKernel with various shapes and configurations."""
    if dim >= len(X_shape): return # Skip if dim is out of bounds
    X = th.randn(th.Size(X_shape), device="cuda")
    Z = interface(X, power, d_tile, dim)
    assert Z.shape == interface_output_shape(X_shape, power, d_tile, dim)
@pytest.mark.parametrize("X_shape,power,d_tile", SHAPES_AND_POWERS_AND_TILES)
def test_interface_small(X_shape, power, d_tile):
    for dim in range(len(X_shape)):
        if X_shape[dim] % d_tile != 0: continue # can't expand this dimension
        _test_interface(X_shape, power, d_tile, dim)
@pytest.mark.parametrize("X_shape,power,d_tile", LARGE_SHAPES_AND_POWERS_AND_TILES)
def test_interface_large(X_shape, power, d_tile):
    _test_interface(X_shape, power, d_tile, -1)

@pytest.mark.parametrize("X_shape,power,d_tile", [
    ([1, 32], 2, 4),
    ([128,32], 3, 1),
    ([8,16], 4, 8),
])
@pytest.mark.parametrize("perm", PERMUTATIONS_OF_STRIDES_UP_TO_4D)
def test_permuted_strides(X_shape, power, d_tile, perm):
    X = th.randn(th.Size(X_shape), device="cuda")
    Z = th.empty(op_output_shape(X_shape, power, d_tile), device="cuda")
    # Create new x tensors with non-standard strides
    X_strided = create_permuted_strides_layout(X, perm)
    Z_strided = create_permuted_strides_layout(Z, perm)
    # the autotuned binding should be invariant to the strides
    binding_autotuned(Z, X, power, d_tile, duplicate_correction=True)
    binding_autotuned(Z_strided, X_strided, power, d_tile, duplicate_correction=True)
    diff(Z, Z_strided, rtol=1e-3, atol=1e-3)

@pytest.mark.parametrize("X_shape, power, d_tile", LARGE_SHAPES_AND_POWERS_AND_TILES)
@pytest.mark.parametrize("duplicate_correction", [True, False])
def test_sympow_autograd_against_reference(X_shape, power, d_tile, duplicate_correction):
    """Test that sympow gradients match the reference implementation."""
    X = th.randn(X_shape, device="cuda", requires_grad=True)
    Z = op(X, power, d_tile, duplicate_correction)
    Zgrad = th.randn_like(Z)
    Z.backward(Zgrad)
    Xgrad = X.grad.clone()
    X.grad = None
    # Run forward and backward with reference implementation
    Z_ref = op_reference(X, power, d_tile, duplicate_correction)
    Z_ref.backward(Zgrad)
    Xgrad_ref = X.grad.clone()
    diff(Xgrad, Xgrad_ref, rtol=1e-3, atol=1e-3)



from torch.utils._pytree import tree_map
from typing import Any, Callable, Optional, Union
from functools import reduce
from vidrial.jit.static.tuples import flatten, unflatten
from vidrial.jit.static.types import Layout
from vidrial.jit.static.shape import Shape, shapify, static_shape
from vidrial.jit.static.stride import Stride, strideify
from vidrial.jit.static.int import Int, _1, _2, _3, _4, _5, _6, _7, _8, _9, _10, _16


def squeeze(thing: Shape | Stride) -> Shape | Stride:
    """Remove all singleton dimensions from a shape or stride."""
    cls = type(thing)
    if len(thing.elements) == 1 and isinstance(thing.elements[0], cls):
        return cls(squeeze(thing.elements[0])) # type: ignore
    else:
        return thing

def merge(*things: Any) -> Any:
    """Merge a list of shapes, strides, or layouts into a single shape, stride, or layout."""
    cls = type(things[0])
    assert all(isinstance(t, cls) for t in things), f"All elements must be of type {cls}"
    if len(things) == 1:
        return squeeze(things[0])
    if cls in (Shape, Stride):
        return squeeze(cls(*things))
    elif cls == Layout:
        return Layout(merge(*(t.shape for t in things)), merge(*(t.stride for t in things)))
    else:
        raise ValueError(f"Cannot merge {cls}")
    
    
def all_factors(n: int) -> list[int]:
    """Return all factors of n."""
    factors = set()
    for i in range(1, int(n**0.5) + 1):
        if n % i == 0:
            factors.add(i)
            factors.add(n // i)
    return sorted(factors)


def common_factors(a: int, b: int) -> list[int]:
    """Return the common factors of a and b."""
    return list(set(all_factors(a)) & set(all_factors(b)))


def all_divides(shape: Shape, splits: int) -> list[Shape]:
    """Return all possible shapes that can be obtained by dividing the shape into splits equal parts."""
    global indents
    if splits == 1:
        return [shape]
    if splits > shape.size():
        return []
    if shape.modes() == 1:
        return [Shape(shape.size() // splits)] if shape.size() % splits == 0 else []

    flattened, res = flatten(shape).to_list(), []
    for factor in common_factors(flattened[0], splits):
        pre_shape = [flattened[0] // factor]
        post_shapes = all_divides(Shape(*flattened[1:]), splits // factor) # type: ignore
        for post_shape in post_shapes:
            res.append(Shape(*pre_shape, *post_shape))
    for i in range(len(res)):
        res[i] = unflatten(res[i], shape)
    return res


class Get:
    def __getitem__(self, idxs: int | tuple[int, ...]) -> Callable:
        if isinstance(idxs, int):
            return lambda l: Layout(shapify(l.shape[idxs]), strideify(l.stride[idxs]))
        def get_layout(l: Layout) -> Layout:
            for idx in idxs:
                l = Get()[idx](l)
            return l
        return get_layout
    
get = Get()



def test_get_1():
    l = Layout(Shape(_1, Shape(_2, _3)), Stride(_4, Stride(_5, _6)))
    l1 = get[0](l)
    l2 = get[1, 1](l)
    assert l1.to_cpp() == "Layout<Shape<Int<1>>, Stride<Int<4>>>"
    assert l2.to_cpp() == "Layout<Shape<Int<3>>, Stride<Int<6>>>"

def test_get_2():
    l = Layout(Shape(_1, Shape(_2, Shape(_3, _4))), Stride(_4, Stride(_5, Stride(_6, _7))))
    l1 = get[1, 1](l)
    l2 = get[1, 1, 1](l)
    assert l1.to_cpp() == "Layout<Shape<Int<3>, Int<4>>, Stride<Int<6>, Int<7>>>"
    assert l2.to_cpp() == "Layout<Shape<Int<4>>, Stride<Int<7>>>"

def test_merge_shape():
    s1 = Shape(_1, _2)
    s2 = Shape(_3)
    s3 = merge(s1, s2)
    assert s3.to_cpp() == "Shape<Shape<Int<1>, Int<2>>, Int<3>>"

def test_merge_stride():
    s1 = Stride(_1, _2)
    s2 = Stride(_3)
    s3 = merge(s1, s2)
    assert s3.to_cpp() == "Stride<Stride<Int<1>, Int<2>>, Int<3>>"

def test_merge_layout_1():
    l1 = Layout(Shape(_1, _2), Stride(_3, _4))
    l2 = Layout(Shape(_5), Stride(_6))
    l3 = merge(l1, l2)
    assert l3.to_cpp() == "Layout<Shape<Shape<Int<1>, Int<2>>, Int<5>>, Stride<Stride<Int<3>, Int<4>>, Int<6>>>"

def test_merge_layout_2():
    l1 = Layout(Shape(_1, _2), Stride(_3, _4))
    l2 = Layout(Shape(_1), Stride(_2))
    l3 = merge(l1, l2)
    assert l3.to_cpp() == "Layout<Shape<Shape<Int<1>, Int<2>>, Int<1>>, Stride<Stride<Int<3>, Int<4>>, Int<2>>>"

def test_all_divides_1():
    s = Shape(_4, Shape(_4, _4), _16)
    res = all_divides(s, 2)
    assert len(res) == 4
    assert any(static_shape(r).to_cpp() == "Shape<Int<2>, Shape<Int<4>, Int<4>>, Int<16>>" for r in res)
    assert any(static_shape(r).to_cpp() == "Shape<Int<4>, Shape<Int<2>, Int<4>>, Int<16>>" for r in res)
    assert any(static_shape(r).to_cpp() == "Shape<Int<4>, Shape<Int<4>, Int<2>>, Int<16>>" for r in res)
    assert any(static_shape(r).to_cpp() == "Shape<Int<4>, Shape<Int<4>, Int<4>>, Int<8>>" for r in res)

def test_all_divides_2():
    s = Shape(_4, Shape(_4, _4), _16)
    res = all_divides(s, 8)
    assert len(res) == 17
    assert len(set(res)) == 17
    assert len(set(r.size() for r in res)) == 1
    assert any(static_shape(r).to_cpp() == "Shape<Int<1>, Shape<Int<2>, Int<4>>, Int<16>>" for r in res)
    assert any(static_shape(r).to_cpp() == "Shape<Int<1>, Shape<Int<4>, Int<2>>, Int<16>>" for r in res)
    assert any(static_shape(r).to_cpp() == "Shape<Int<1>, Shape<Int<4>, Int<4>>, Int<8>>" for r in res)
    assert any(static_shape(r).to_cpp() == "Shape<Int<2>, Shape<Int<2>, Int<2>>, Int<16>>" for r in res)
    assert any(static_shape(r).to_cpp() == "Shape<Int<2>, Shape<Int<2>, Int<4>>, Int<8>>" for r in res)
    assert any(static_shape(r).to_cpp() == "Shape<Int<2>, Shape<Int<4>, Int<2>>, Int<8>>" for r in res)
    assert any(static_shape(r).to_cpp() == "Shape<Int<4>, Shape<Int<1>, Int<2>>, Int<16>>" for r in res)
    assert any(static_shape(r).to_cpp() == "Shape<Int<4>, Shape<Int<2>, Int<1>>, Int<16>>" for r in res)
    assert any(static_shape(r).to_cpp() == "Shape<Int<4>, Shape<Int<2>, Int<2>>, Int<8>>" for r in res)
    assert any(static_shape(r).to_cpp() == "Shape<Int<4>, Shape<Int<4>, Int<1>>, Int<8>>" for r in res)
    assert any(static_shape(r).to_cpp() == "Shape<Int<4>, Shape<Int<1>, Int<4>>, Int<8>>" for r in res)
    assert any(static_shape(r).to_cpp() == "Shape<Int<4>, Shape<Int<2>, Int<4>>, Int<4>>" for r in res)
    assert any(static_shape(r).to_cpp() == "Shape<Int<4>, Shape<Int<4>, Int<2>>, Int<4>>" for r in res)
    assert any(static_shape(r).to_cpp() == "Shape<Int<4>, Shape<Int<4>, Int<4>>, Int<2>>" for r in res)



from collections.abc import Iterable
from functools import wraps
import inspect
import torch

def sanitize_kwargs(fn):
    """
    Sanitizes kwargs by removing any that are not in the function signature.
    """
    @wraps(fn)
    def wrapper(**kwargs):
        sig = inspect.signature(fn)
        valid_kwargs = {k: v for k, v in kwargs.items() if k in sig.parameters}
        return fn(**valid_kwargs)
    return wrapper

def tensors_to_ones_like(items):
    """Helper function to convert a tensor or iterable/dict of tensors to ones of the same shape."""
    if isinstance(items, torch.Tensor):
        return torch.ones_like(items)
    elif isinstance(items, dict):
        return {k: tensors_to_ones_like(v) for k, v in items.items() if isinstance(v, torch.Tensor)}
    else:
        return tuple(tensors_to_ones_like(item) for item in items if isinstance(item, torch.Tensor))

def get_compiled_version(fn, inputs, direction, warmup=3, compile=True):
    """Takes a function and args and returns compiled versions for fwd, bwd, and fwd+bwd passes.

    Args:
        fn: Function to compile
        inputs: A dict, keyword arguments to pass to fn

    Returns:
        Tuple of (fwd_fn, bwd_fn, fwd_bwd_fn)
    """
    # Run forward pass to identify output shapes and check for mutations
    # Define functions
    def fwd():
        with torch.no_grad():
            return sanitize_kwargs(fn)(**inputs)
    torch._dynamo.config.compiled_autograd = True
    outputs = sanitize_kwargs(fn)(**inputs)
    grads = tensors_to_ones_like(outputs)
    def bwd():
        nonlocal grads, outputs
        tuples = [(o, g) for o, g in zip(outputs, grads) if o.requires_grad]
        outputs, grads = [t[0] for t in tuples], [t[1] for t in tuples]
        torch.autograd.backward(outputs, grad_tensors=grads, retain_graph=True)
    torch._dynamo.config.compiled_autograd = False
    def fwd_bwd():
        nonlocal grads
        outputs = sanitize_kwargs(fn)(**inputs)
        tuples = [(o, g) for o, g in zip(outputs, grads) if o.requires_grad]
        outputs, grads = [t[0] for t in tuples], [t[1] for t in tuples]
        torch.autograd.backward(outputs, grad_tensors=grads)
    # Compile functions
    if direction == 'fwd':
        compiled_fn = torch.compile(fwd, dynamic=False) if compile else fwd
    elif direction == 'bwd':
        compiled_fn = torch.compile(bwd, dynamic=False) if compile else bwd
    elif direction == 'fwd+bwd':
        compiled_fn = torch.compile(fwd_bwd, dynamic=False) if compile else fwd_bwd
    # Warmup passes
    for _ in range(warmup):
        compiled_fn()
    # Return compiled functions
    return compiled_fn



def check_tensors_unchanged(tensor1, tensor2, prefix=''):
    assert type(tensor1) == type(tensor2), f'Mismatch in inputs: {type(tensor1)=}, {type(tensor2)=}'
    if isinstance(tensor1, torch.Tensor):
        assert tensor1.shape == tensor2.shape, f'{prefix}Functions must not mutate their inputs: Tensor shapes were modified'
        assert tensor1.dtype == tensor2.dtype, f'{prefix}Functions must not mutate their inputs: Tensor dtypes were modified'
        assert tensor1.device == tensor2.device, f'{prefix}Functions must not mutate their inputs: Tensor devices were modified'
        assert tensor1.stride() == tensor2.stride(), f'{prefix}Functions must not mutate their inputs: Tensor strides were modified'
    elif isinstance(tensor1, Iterable):
        for t, o in zip(tensor1, tensor2, strict=False):
            check_tensors_unchanged(t, o, prefix)


def wrap_with_timer(fn, n=10, warmup=3):
    """Takes a function and returns a function that calls it n times and returns the total time."""
    def timed_fn(*args, **kwargs):
        torch.cuda.synchronize()
        for _ in range(warmup):
            fn(*args, **kwargs)

        cache = torch.empty(int(256e6), dtype=torch.int8, device='cuda')
        def flush_cache():
            cache.zero_()

        start_events = [torch.cuda.Event(enable_timing=True) for _ in range(n)]
        end_events = [torch.cuda.Event(enable_timing=True) for _ in range(n)]
        for i in range(n):
            flush_cache()
            torch.cuda._sleep(5_000_000_0)
            start_events[i].record()
            out = fn(*args, **kwargs)
            end_events[i].record()
        torch.cuda.synchronize()
        times = [s.elapsed_time(e) for s, e in zip(start_events, end_events)]
        total_time = sum(times)
        return out, total_time
    return timed_fn

def estimate_runtime(fn, *args, num1=10, num2=50, **kwargs):
    """Takes a function and returns a an estimate of time per iteration."""
    timed_fn_1 = wrap_with_timer(fn, num1)
    timed_fn_2 = wrap_with_timer(fn, num2)

    _, t1 = timed_fn_1(*args, **kwargs)
    _, t2 = timed_fn_2(*args, **kwargs)

    return (t2 - t1) / (num2 - num1)


def benchmark_speed(direction, fn, create_inputs, create_inputs_kwargs, num1=10, num2=30, warmup=3, compile=True):
    """Measure speed of a function implementation.
    
    Args:
        direction: str. One of 'fwd', 'bwd', or 'fwd+bwd' to measure forward pass,
            backward pass, or combined forward+backward pass timing
        fn: Function to benchmark
        create_inputs: Function that creates input tensors for fn
        **kw: Keyword arguments passed to create_inputs() to configure the test case
            
    Returns:
        float: Time in milliseconds per iteration
    """
    inputs = create_inputs(**create_inputs_kwargs, requires_grad=True)
    fn = get_compiled_version(fn, inputs, direction=direction, warmup=warmup, compile=compile)
    return estimate_runtime(fn, num1=num1, num2=num2)



from vidrial.kernels.sympow.dimensions import sympow_dim, sympow_shape

def problem_shape(A_shape, B_shape, C_shape, dim, power, d_tile):
    ndim = len(A_shape)
    batches, M, N, K, d = tuple(C_shape[:-2]), C_shape[-2], C_shape[-1], B_shape[-2], A_shape[dim]
    assert len(batches) >= 1
    D = sympow_dim(d, power, d_tile)
    eA_shape = sympow_shape(A_shape, power, d_tile, dim)
    assert dim % ndim - ndim in [-1,-2], f"expand_dim={dim} must corrsepond M or K (-2 or -1)"
    assert eA_shape == batches + (M, K), f"eA_shape={eA_shape} != batches + (M, K)={batches + (M, K)}"
    tmp = list(eA_shape)
    tmp[dim] = d
    assert A_shape == tuple(tmp)
    assert B_shape == batches + (K, N)
    assert C_shape == batches + (M, N)
    return batches, M, N, K, d, D

def op_output_shape(A_shape, B_shape, expand_dim, power, d_tile) -> tuple[int, int, int]:
    ndim = len(A_shape)
    batches, K, N = tuple(B_shape[:-2]), B_shape[-2], B_shape[-1]
    d = A_shape[expand_dim]
    expand_M = expand_dim%ndim-ndim == -2
    M = sympow_dim(d, power, d_tile) if expand_M  else A_shape[-2]
    C_shape = batches+(M, N)
    problem_shape(A_shape, B_shape, C_shape, expand_dim, power, d_tile) # Check that the shape is valid
    return C_shape



from typing import Callable, Any, Iterable
from tqdm import tqdm
from concurrent.futures import ThreadPoolExecutor, as_completed
import inspect
import time
import random
import logging
from vidrial.jit import settings
from vidrial.jit.timingcache import ConfigTimingCache
from vidrial.jit.timer import timeit
from vidrial.py_utils.test_utils import exact_copy

logger = logging.getLogger(__name__)

def get_fn_name(fn: Callable) -> str:
    return f"{fn.__module__}.{fn.__qualname__}".replace("<locals>.", "")

def parallel_warmup(fn: Callable,
                    args: dict,
                    configs: list[dict],
                    allow_failure: bool,
                    max_workers: int,
                    verbose: bool,
                    name: str | None = None) -> list[dict]:
    """
    Warmup a given function in parallel (i.e., call it once for each config). The keys of `args_dict`
    and the keys of each config together must completely specify the arguments of the function.

    :param fn: The function to warmup.
    :param args: A dictionary of "fixed" arguments to the function
    :param configs: A list of dictionaries of containing additional arguments to the function.
    :param allow_failure: If True, the tuner will not raise an error if at least one of the configurations succeeds to run.
    :param max_workers: The maximum number of workers to use for the warmup.
    :param verbose: If True, the tuner will print progress information.
    :param name: optional name of the kernel to be used for logging.

    :return: A list of configurations that succeeded to run.
    """
    def _run(config):
        fn(*args.values(), **config)
        return config

    fn_name = name or get_fn_name(fn)
    
    errors, success_configs, start_time = [], [], time.time()
    with ThreadPoolExecutor(max_workers=min(max_workers, len(configs))) as executor:
        futures = [executor.submit(_run, config) for config in configs]
        with tqdm(total=len(configs), desc=f"Warming up for {fn_name}", disable=not verbose) as pbar:
            for future in as_completed(futures):
                try:
                    config = future.result()
                    success_configs.append(config)
                except Exception as e:
                    logger.debug(f"Error during warmup: {e}")
                    errors.append(str(e))
                pbar.update(1)
    logger.debug(f"{len(success_configs)=}, {len(configs)=}, {allow_failure=}")
    if len(success_configs) < len(configs) and ((not allow_failure) or len(success_configs) == 0):
        error_msg = f"{len(configs) - len(success_configs)} out of {len(configs)} warmup runs failed for {fn_name}:\n"
        for i, err in enumerate(errors):
            error_msg += f"- Config {i}: {err}\n"
        logger.error(error_msg)
        raise RuntimeError(error_msg)

    logger.debug(f"Warmup completed for {fn_name}: {len(success_configs)} successful, {len(errors)} failed in {time.time() - start_time:.2f} seconds")
    return success_configs


def kerneltune(
    fn: Callable,
    args: dict,
    configs: list[dict] | Callable, # type: ignore
    num: int = 10,
    max_workers: int = 8,
    no_side_effect: bool = True,
    verbose: bool = False,
    allow_failure: bool = True,
    name: str | None = None):
    """
    Tunes a kernel by test-running different configurations, returns timing results in the following format:
    [
        {
            "config": {...}, # The configuration that was tested
            "time": float,   # Average time of the run
            "std": float,    # Standard deviation of the time
            "n": int,        # Number of runs incurred
        },
        ...
    ]

    Example:
        def add_one_kernel(X, Y, d0_tile, d1_tile, thread_num):
            runtime = jit(...)
            runtime(X, Y)
            return Y

        timings = kerneltune(
            fn=add_one_kernel,
            args={'X': X, 'Y': Y},
            configs=[
                {'d0_tile': 16, 'd1_tile': 16, 'thread_num': 16},
                {'d0_tile': 32, 'd1_tile': 32, 'thread_num': 32},
                {'d0_tile': 64, 'd1_tile': 64, 'thread_num': 64},
            ],
            no_side_effects=True,
        )


    :note: When all the configurations are evaluated, the kernel will run multiple times. This means that whatever value the kernel updates will be updated multiple times. To avoid this undesired behavior, you can use the `reset_to_zero` argument, which resets the value of the provided tensor to `zero` before running any configuration.
        
    :note: Unlike triton.autotune, we use standard nvcc compilation tool chain to compile the kernel. This means that we'll spin up multiple processes to compile the kernels for different configurations first before benchmarking.

    :param fn: the runnable to tune
    :param configs: List of config dictionaries, or functions that return a list of config dictionaries
    :param num: number of times to run the kernel for each configuration
    :param max_workers: number of workers to use for the warmup
    :param no_side_effect: if True, the tuner will restore the values of the provided argument names after evaluating any configs.
    :param verbose: if True, the tuner will print progress information.
    :param allow_failure: if True, the tuner will not raise an error if at least one of the configurations succeeds to run.
    :param name: optional name of the kernel to be used for logging.
    """
    assert num >= 3, "num must be at least 3"
    num1 = num // 3
    num2 = num - num1
    fn_name = name or get_fn_name(fn)

    def _bench(config: dict) -> dict:
        args_copy = exact_copy(args)
        avg_time, std_time = timeit(lambda: fn(*args_copy.values(), **config), num1=num1, num2=num2)
        logger.debug(f"Time for {fn_name} with config {config}: {avg_time:.2f} ± {std_time:.2f} seconds")
        return {'config': config, 'time': avg_time, 'std': std_time, 'n': num1 + num2}

    if isinstance(configs, Callable):
        configs: list[dict] = configs(args.values())
    else:
        configs = configs
    bench_start = time.time()
    configs = parallel_warmup(fn, args, configs, allow_failure, max_workers, verbose, fn_name)
    if settings.precompile:
        return [{'config': config, 'time': 0, 'std': 0, 'n': 0} for config in configs]
    timings = [_bench(config=config) for config in (tqdm(configs, desc="Benchmarking") if verbose else configs)]
    bench_end = time.time()
    bench_time = bench_end - bench_start
    logger.debug(f"Tuning time for {fn_name}: {bench_time:.2f} seconds")
    timings = sorted(timings, key=lambda x: x['time'])
    config = timings[0]['config']
    logger.debug(f"Best config for {fn_name}: {config}")
    return timings
    

def tune_and_update(fn: Callable,
                    args: Iterable[Any],
                    configs: list[dict] | Callable,
                    cache: ConfigTimingCache,
                    num: int = 10,
                    max_workers: int = 8,
                    no_side_effect: bool = True,
                    verbose: bool = False,
                    allow_failure: bool = True) -> list[dict]:
    """ Tunes a function and updates the function with the collected timings.
    """
    if not isinstance(args, dict):
        arg_names = list(inspect.signature(fn).parameters.keys())
        args = dict(zip(arg_names, args))
    timings = kerneltune(fn, args, configs, num, max_workers, no_side_effect, verbose, allow_failure)
    if not settings.precompile: # do not update the cache if precompiling, all the timings will be wrong
        cache.update(args, timings)
    else:
        logger.debug(f"Precompiling mode, not updating the cache")
    return timings

def sample_configs(configs: list[dict], n: int = 10) -> list[dict]:
    """Sample a random subset of the configurations.
    """
    random.seed(42)
    return random.sample(configs, min(n, len(configs)))


from vidrial.jit.static import Shape, Stride, Layout, UniversalCopy, CopyAtom, CudaDType, Int, compact_major, _32, _64, _128, _256, _8

def test_basic_shapes():
    simple_shape = Shape(_32, _32)
    assert simple_shape.to_cpp() == "Shape<Int<32>, Int<32>>"

def test_nested_shapes():
    nested_shape = Shape(Shape(_32, _32), _64)
    assert nested_shape.to_cpp() == "Shape<Shape<Int<32>, Int<32>>, Int<64>>"

def test_complex_nested_shapes():
    complex_shape = Shape(Shape(_32, Shape(_64, _128)), Shape(_256, _32))
    assert complex_shape.to_cpp() == "Shape<Shape<Int<32>, Shape<Int<64>, Int<128>>>, Shape<Int<256>, Int<32>>>"

def test_basic_layouts():
    simple_shape = Shape(_32, _32)
    simple_layout = Layout(simple_shape)
    assert simple_layout.to_cpp() == "Layout<Shape<Int<32>, Int<32>>, Stride<Int<1>, Int<32>>>"

def test_right_layout():
    simple_shape = Shape(_32, _32)
    simple_layout = Layout(simple_shape, compact_major(simple_shape, left=False))
    assert simple_layout.to_cpp() == "Layout<Shape<Int<32>, Int<32>>, Stride<Int<32>, Int<1>>>"

def test_layouts_with_strides():
    simple_shape = Shape(_32, _32)
    custom_stride = Stride(_64, _32)
    layout_with_stride = Layout(simple_shape, custom_stride)
    assert layout_with_stride.to_cpp() == "Layout<Shape<Int<32>, Int<32>>, Stride<Int<64>, Int<32>>>"

def test_nested_layouts():
    simple_shape = Shape(_32, _32)
    nested_layout = Layout(Shape(simple_shape, _64))
    assert nested_layout.to_cpp() == "Layout<Shape<Shape<Int<32>, Int<32>>, Int<64>>, Stride<Stride<Int<1>, Int<32>>, Int<1024>>>"

def test_basic_traits():
    float_trait = UniversalCopy(CudaDType.FLOAT)
    assert float_trait.to_cpp() == "UniversalCopy<float>"

def test_copy_atoms():
    float_trait = UniversalCopy(CudaDType.FLOAT)
    float_copy_atom = CopyAtom(float_trait)
    assert float_copy_atom.to_cpp() == "Copy_Atom<UniversalCopy<float>, float>"

def test_complex_structure():
    complex_structure = Layout(
        Shape(
            Shape(_32, _64),
            Shape(_128, _256)
        ),
        Stride(
            _32,
            Stride(_64, _128)
        )
    )
    assert complex_structure.to_cpp() == "Layout<Shape<Shape<Int<32>, Int<64>>, Shape<Int<128>, Int<256>>>, Stride<Int<32>, Stride<Int<64>, Int<128>>>>"

def test_complex_trait_structure():
    complex_trait_structure = CopyAtom(
        UniversalCopy(CudaDType.UINT8)
    )
    assert complex_trait_structure.to_cpp() == "Copy_Atom<UniversalCopy<uint8_t>, uint8_t>"


def test_compact_major():
    shape = Shape(_32, _32)
    stride_left = compact_major(shape)
    assert stride_left.to_cpp() == "Stride<Int<1>, Int<32>>"
    stride_right = compact_major(shape, left=False)
    assert stride_right.to_cpp() == "Stride<Int<32>, Int<1>>"
    shape1 = Shape(Shape(Shape(_8, _8), Int(3)), _8)
    layout_left = Layout(shape1)
    assert layout_left.to_cpp() == "Layout<Shape<Shape<Shape<Int<8>, Int<8>>, Int<3>>, Int<8>>, Stride<Stride<Stride<Int<1>, Int<8>>, Int<64>>, Int<192>>>"


from dataclasses import dataclass
from typing import List, Dict, Any
import torch as th
from vidrial.jit.jit import render, jit
from vidrial.jit.binding import make_binding
from vidrial.jit.timingcache import ConfigTimingCache

# ------------------- Source Code -------------------
@dataclass
class SourceCode:
    d0: int
    d1: int
    tile_d0: int
    tile_d1: int
    thread_num: int
    
    @property
    def template(self) -> str:
        return """
#include "add_one/add_one_kernels.cuh"
#include "copy/copy_kernels.cuh"
using namespace cute;
using namespace vidrial;

extern "C" {
    void launch(float* Vptr) {
        using SlabShape = Shape<Int<{d0}>, Int<{d1}>>;
        using TileShape = Shape<Int<{tile_d0}>, Int<{tile_d1}>>;
        using ASlab = Layout<SlabShape>;
        launch_add_one_inplace<float, {thread_num}, SlabShape, TileShape, ASlab>(Vptr);
    }
}"""

    def __str__(self) -> str:
        return render(self.template, self.__dict__)


@dataclass
class BindingCfg:
    d0: int
    d1: int
    tile_d0: int
    tile_d1: int
    thread_num: int
    
    @classmethod
    def from_args(cls, X, tile_d0, tile_d1, thread_num):
        return cls(
            d0=X.shape[0],
            d1=X.shape[1],
            tile_d0=tile_d0,
            tile_d1=tile_d1,
            thread_num=thread_num)

    @property
    def source(self):
        return SourceCode(
            d0=self.d0,
            d1=self.d1,
            tile_d0=self.tile_d0,
            tile_d1=self.tile_d1,
            thread_num=self.thread_num)


def binding(X: th.Tensor, tile_d0: int, tile_d1: int, thread_num: int):
    cfg = BindingCfg.from_args(X, tile_d0, tile_d1, thread_num)
    jit(name = "add_one", code = str(cfg.source))(X)


# ---------------------- Autotune --------------------------------

def make_add_one_configurator(num_range: List[int]):
    def add_one_configurator(args: dict) -> List[dict[str, Any]]:
        X = args['X']
        configs = []
        d0, d1 = X.shape
        for thread_num in num_range:
            for tile_d0 in num_range:
                for tile_d1 in num_range:
                    if tile_d0 > d0 or tile_d1 > d1:
                        continue
                    configs.append({
                        'thread_num': thread_num,
                        'tile_d0': tile_d0,
                        'tile_d1': tile_d1,
                    })

        return configs
    return add_one_configurator


def hash_fn(args: dict) -> str:
    X = args['X']
    return f"{X.shape[0]}_{X.shape[1]}_{X.dtype}"

cache = ConfigTimingCache('add_one', hash_fn)
add_one_kernel = make_binding(cache=cache, sweep=make_add_one_configurator([1, 4, 32]))(binding)


if __name__ == '__main__':
    input = th.ones((64,64), dtype=th.float32, device="cuda")
    add_one_kernel(input)



from dataclasses import dataclass
from typing import Any
import torch as th
from vidrial.jit.static.types import Shape, Layout, Int
from vidrial.jit.static.util import layout_from_shape_stride, torch_dtype_to_c
from vidrial.jit.timingcache import ConfigTimingCache
from vidrial.jit.binding import make_binding
from vidrial.jit.jit import jit, render
from vidrial.kernels.mma_configurator import base_configurator
from vidrial.kernels.mma.dimensions import problem_shape


# ------------------- Source Code -------------------

@dataclass
class SourceCode:
    T: str
    Atom: str
    MNKTileShape: Shape
    MNKAtomPlacement: Shape
    GASlab: Layout
    GBSlab: Layout
    GCSlab: Layout
    @property
    def template(self):
        return """
#include <cuda.h>
#include <cuda_runtime.h>
#include <iostream>
#include "cute/tensor.hpp"
#include "kernels/mma/kernel.cuh"
#include "kernels/mma/mma_cfg.cuh"

using namespace cute;
using namespace vidrial;

extern "C" int launch(void* __raw_A, void* __raw_B, void* __raw_C) {
    using T = {T};
    using Atom = MMA_Atom<{Atom}>;
    using MNKTileShape = decltype(static_tree_cast<int64_t>({MNKTileShape}{}));
    using MNKAtomPlacement = decltype(static_tree_cast<int64_t>({MNKAtomPlacement}{}));
    using GASlab = decltype(static_tree_cast<int64_t>({GASlab}{}));
    using GBSlab = decltype(static_tree_cast<int64_t>({GBSlab}{}));
    using GCSlab = decltype(static_tree_cast<int64_t>({GCSlab}{}));
    using Cfg = MmaKernelCfg<T, Atom, MNKTileShape, MNKAtomPlacement, GASlab, GBSlab, GCSlab>;
    auto A = reinterpret_cast<T*>(__raw_A);
    auto B = reinterpret_cast<T*>(__raw_B);
    auto C = reinterpret_cast<T*>(__raw_C);
    return launch_tiled_mma_kernel(Cfg{}, A, B, C);
}"""
    def __str__(self):
        return render(self.template, self.__dict__)


# ------------------- Binding -------------------

@dataclass
class BindingCfg:
    P: int # batch dimension
    M: int
    N: int
    K: int
    A_shape: tuple[int, ...]
    B_shape: tuple[int, ...]
    C_shape: tuple[int, ...]
    A_stride: tuple[int, ...]
    B_stride: tuple[int, ...]
    C_stride: tuple[int, ...]
    Atom: str
    MNKTileShape: tuple[int, ...]
    MNKAtomPlacement: tuple[int, ...]
    smempipe: int
    regpipe: int
    use_ldsm: bool
    swizzle: int
    dtype: th.dtype
    @classmethod
    def from_args(cls, A, B, C, Atom, MNKTileShape, MNKAtomPlacement, smempipe, regpipe, use_ldsm, swizzle):
        assert A.device.type == B.device.type == C.device.type == 'cuda', f"Invalid {A.device}, {B.device}, {C.device}."
        assert A.dtype == B.dtype == C.dtype, f"Invalid {A.dtype}, {B.dtype}, {C.dtype}. Kernel currently assumes all input tensors have the same dtype"
        assert smempipe == 1, "currently only default smempipe is supported"
        assert regpipe == 1, "currently only default regpipe is supported"
        assert use_ldsm is True, "currently only default use_ldsm is supported"
        assert swizzle == 1, "currently only default swizzle is supported"  
        P, M, N, K = problem_shape(A.shape, B.shape, C.shape)
        return cls(P, M, N, K, A.shape, B.shape, C.shape, A.stride(), B.stride(), C.stride(), Atom, MNKTileShape, MNKAtomPlacement, smempipe, regpipe, use_ldsm, swizzle, A.dtype)
    @property
    def source(self):
        """ Conversion between the different variable name and shape conventions
                Python                        c++
            A_shape=[P,M,K]      ->      GASlabShape=[M,K,P]
            B_shape=[P,K,N]      ->      GBSlabShape=[N,K,P]
            C_shape=[P,M,N]      ->      GCSlabShape=[M,N,P]
        """
        GASlab = layout_from_shape_stride(self.A_shape, self.A_stride, (1, 2, 0))
        GBSlab = layout_from_shape_stride(self.B_shape, self.B_stride, (2, 1, 0))
        GCSlab = layout_from_shape_stride(self.C_shape, self.C_stride, (1, 2, 0))
        return SourceCode(
            T=torch_dtype_to_c(self.dtype),
            Atom=self.Atom,
            MNKTileShape=Shape(*[Int(x) for x in self.MNKTileShape]),
            MNKAtomPlacement=Shape(*[Int(x) for x in self.MNKAtomPlacement]),
            GASlab=GASlab,
            GBSlab=GBSlab,
            GCSlab=GCSlab,
        )

# ---------------------- Autotune --------------------------------

def make_mma_configurator(smempipe=None, regpipe=None, use_ldsm=None, swizzle=None):
    def mma_configurator(args: dict) -> list[dict[str, Any]]:
        A, B, C = args['A'], args['B'], args['C']
        P, M, N, K = problem_shape(A.shape, B.shape, C.shape)
        return base_configurator(MNKP=(M, N, K, P), dtype=A.dtype, acc_dtype=th.float32, smempipe=smempipe, regpipe=regpipe, use_ldsm=use_ldsm, swizzle=swizzle)
    return mma_configurator

def hash_fn(args: dict) -> str:
    A, B, C = args['A'], args['B'], args['C']
    P, M, N, K = problem_shape(A.shape, B.shape, C.shape)
    A_major = 'K_major' if A.stride(-1) == 1 else 'M_major'
    B_major = 'N_major' if B.stride(-1) == 1 else 'K_major'
    C_major = 'N_major' if C.stride(-1) == 1 else 'M_major'
    return f"{A.dtype}-{B.dtype}-{C.dtype}-M_mod128_{M % 128}-N_mod128_{N % 128}-K_mod128_{K % 128}-{A_major}-{B_major}-{C_major}"

@make_binding(cache=ConfigTimingCache('mma', hash_fn),
          sweep=make_mma_configurator(smempipe=1, regpipe=1, use_ldsm=True, swizzle=1))
def binding(A, B, C, Atom, MNKTileShape, MNKAtomPlacement, smempipe, regpipe, use_ldsm, swizzle):
    binding_cfg = BindingCfg.from_args(A, B, C, Atom, MNKTileShape, MNKAtomPlacement, smempipe, regpipe, use_ldsm, swizzle)
    jit(name = "mma",
        code = str(binding_cfg.source)
    )(A, B, C)



import torch
from unittest.mock import patch
from vidrial.jit.jit import jit
from vidrial.jit.compiler import compile

def test_jit(get_add_one_kernel_code, tmp_path):
    with patch('vidrial.jit.jit.compile', wraps=compile) as mock_compile:
        X = torch.randn(64, 64, device="cuda")
        Y = torch.empty_like(X)
        code = get_add_one_kernel_code(X, Y)
        runtime = jit("test_add_one_kernel", code, root=str(tmp_path))
        runtime(X, Y)
        torch.testing.assert_close(X + 1, Y)
        mock_compile.assert_called_once()

        # Now call it second time, it should be cached
        mock_compile.reset_mock()
        runtime = jit("test_add_one_kernel", code, root=str(tmp_path))
        runtime(X, Y)
        assert mock_compile.call_count == 0



from torch.utils._pytree import register_pytree_node
from vidrial.jit.static.int import Int
from vidrial.jit.static.tuples import Tuple, flatten, unflatten

class Stride(Tuple):
    """Represents a stride type, similar to Shape but for strides. A stride can be
    constructed from a (mixed) list of ints, strides, or Ints, but once constructed,
    all elements are either Ints or Strides."""
    name = "Stride"


def strideify(stride: Stride | Int) -> Stride:
    if isinstance(stride, Int):
        return Stride(stride)
    return stride


register_pytree_node(Stride, lambda t: (list(flatten(t)), t), unflatten)


import pytest
import torch as th
from math import sqrt
from vidrial.kernels.sympow.dimensions import sympow_dim
from vidrial.jit.static import Shape, Layout, Int, Tuple
from vidrial.kernels.mma_sympow_bwd.binding import binding
from vidrial.kernels.mma_sympow_bwd.interface import interface, interface_reference
from vidrial.py_utils.test_utils import diff
from vidrial.jit import settings, PickAny

def test_binding():
    d, power, d_tile = 64, 2, 8
    D = sympow_dim(d, power, d_tile)
    M, N, K, batches = D, 16, 32, (1, 1)
    A = th.randn(batches + (M, K), device="cuda", dtype=th.float16)
    B = th.randn(batches + (K, N), device="cuda", dtype=th.float16)
    c = th.randn(batches + (d, N), device="cuda", dtype=th.float16)
    c_dot = th.empty(batches + (d, N), device="cuda", dtype=th.float16)

    binding(
        A, B, c, c_dot, -2, power, d_tile, 1.0,
        duplicate_correction=True,
        MNKTileShape=Shape(Int(64), Int(16), Int(32)), 
        MNKAtomPlacement=Shape(Int(1), Int(1), Int(1)), 
        Atom="SM80_16x8x8_F16F16F16F16_TN", 
        smempipe=1, 
        regpipe=2, 
        use_ldsm=True, 
        swizzle=0,
        SmemAccI=Tuple(*[Int(1) for i in range(power)])
    )

def test_call_interface_expand_M():
    d, power, d_tile = 64, 2, 8
    D = sympow_dim(d, power, d_tile)
    M, N, K, P = D, 16, 32, 8
    A = th.randn((P, M, K), device="cuda")
    B = th.randn((P, K, N), device="cuda")
    c = th.randn((P, d, N), device="cuda")
    c_dot = interface(A, B, c, -2, power, d_tile)

def test_call_interface_expand_N():
    d, power, d_tile = 64, 2, 8
    D = sympow_dim(d, power, d_tile)
    M, N, K, P = 16, D, 32, 8
    A = th.randn((P, M, K), device="cuda")
    B = th.randn((P, K, N), device="cuda")
    c = th.randn((P, M, d), device="cuda")
    c_dot = interface(A, B, c, -1, power, d_tile)

@pytest.mark.parametrize("d, power, d_tile, N, K, batches, dtype", [
    (4, 2, 4, 8, 4, (1,), th.float32),
    (64, 2, 8, 64, 128, (4,2), th.float32),
    (64, 2, 1, 64, 1, (128,2,1), th.float32),
    (4, 2, 4, 8, 8, (1,), th.bfloat16),
    (64, 2, 8, 64, 128, (4,2), th.bfloat16),
    (64, 2, 1, 64, 1, (128,2,1), th.bfloat16),
    (4, 2, 4, 8, 8, (1,), th.float16),
    (64, 2, 8, 64, 128, (4,2), th.float16),
    (64, 2, 1, 64, 1, (128,2,1), th.float16)
])
def test_interface_matches_reference(d, power, d_tile, N, K, batches, dtype):
    M = sympow_dim(d, power, d_tile)
    th.manual_seed(42)
    A = th.randn(batches + (M, K), device="cuda", dtype=dtype) / sqrt(M + K)
    B = th.randn(batches + (K, N), device="cuda", dtype=dtype) / sqrt(K + N)
    c = th.randn(batches + (d, N), device="cuda", dtype=dtype) / sqrt(d + N)
    with settings.set(policy=PickAny, allow_failure=False, verbose=True):
        c_dot = interface(A, B, c, -2, power, d_tile)
    c_dot_ref = interface_reference(A, B, c, -2, power, d_tile)
    tol = 1e-1 if dtype == th.bfloat16 else 1e-3
    print(tol, dtype)
    diff(c_dot, c_dot_ref, atol=tol, rtol=tol, verbose=True)



from typing import Callable
from contextlib import contextmanager
import copy
from dataclasses import dataclass
import logging

logger = logging.getLogger(__name__)

class ModePickAny:
    __repr__ = lambda self: "PickAny"

class ModePickBest:
    __repr__ = lambda self: "PickBest"

PickAny = ModePickAny()
PickBest = ModePickBest()

@dataclass
class Settings:
    policy: ModePickAny | ModePickBest = PickAny
    max_workers: int = 32
    allow_failure: bool = True
    verbose: bool = False
    max_configs: int = 512
    precompile: bool = False
    plot_timings: bool = False

    @contextmanager
    def set(self,
            policy: ModePickAny | ModePickBest | None = None,
            max_workers: int | None = None,
            allow_failure: bool | None = None,
            verbose: bool | None = None,
            max_configs: int | None = None,
            precompile: bool | None = None,
            plot_timings: bool | None = None):
        old_settings = copy.copy(self)
        if policy is not None:
            logger.info(f"Setting policy to {policy}")
            assert isinstance(policy, (ModePickAny, ModePickBest))
            self.policy = policy
        if max_workers is not None:
            logger.info(f"Setting max_workers to {max_workers}")
            assert isinstance(max_workers, int)
            self.max_workers = max_workers
        if allow_failure is not None:
            logger.info(f"Setting allow_failure to {allow_failure}")
            assert isinstance(allow_failure, bool)
            self.allow_failure = allow_failure
        if verbose is not None:
            logger.info(f"Setting verbose to {verbose}")
            assert isinstance(verbose, bool)
            self.verbose = verbose
        if max_configs is not None:
            logger.info(f"Setting max_configs to {max_configs}")
            assert isinstance(max_configs, int)
            self.max_configs = max_configs
        if precompile is not None:
            logger.info(f"Setting precompile to {precompile}")
            assert isinstance(precompile, bool)
            self.precompile = precompile
        if plot_timings is not None:
            logger.info(f"Setting plot_timings to {plot_timings}")
            if self.policy != PickBest and plot_timings:
                logger.warning("plot_timings is only effective in PickBest mode")
            assert isinstance(plot_timings, bool)
            self.plot_timings = plot_timings
        try:
            yield
        finally:
            if self.precompile:
                import vidrial.jit.jit as jit # Lazy import to avoid circular dependency
                jit.precompile()
            logger.info(f"Restoring old settings")
            self.policy = old_settings.policy
            self.max_workers = old_settings.max_workers
            self.allow_failure = old_settings.allow_failure
            self.verbose = old_settings.verbose
            self.max_configs = old_settings.max_configs
            self.precompile = old_settings.precompile
            self.plot_timings = old_settings.plot_timings

# Global settings instance
settings = Settings()



from typing import Tuple

def problem_shape(A_shape, B_shape, C_shape=None) -> Tuple[int, int, int, int]:
    assert len(A_shape) == len(B_shape) == 3
    if C_shape is not None: assert len(A_shape) == len(C_shape)
    P, M, N, K = A_shape[0], A_shape[-2], B_shape[-1], A_shape[-1]
    assert A_shape == (P, M, K)
    assert B_shape == (P, K, N)
    if C_shape is not None: assert C_shape == (P, M, N)
    return P, M, N, K




import torch
from vidrial.jit.tuner import kerneltune, tune_and_update, get_fn_name
from vidrial.jit.timingcache import ConfigTimingCache


def test_tuner(add_one_kernel):
    X = torch.randn(1024, 1024, device="cuda")
    Y = torch.empty_like(X)
    timeings = kerneltune(
        add_one_kernel,
        {'X': X, 'Y': Y},
        configs=[{'d0_tile': 16, 'd1_tile': 16, 'thread_num': 32}, {'d0_tile': 32, 'd1_tile': 32, 'thread_num': 32}, {'d0_tile': 64, 'd1_tile': 64, 'thread_num': 64}],
        num=10,
        no_side_effect=True)
    assert len(timeings) == 3
    assert any(timing['config'] == {'d0_tile': 16, 'd1_tile': 16, 'thread_num': 32} for timing in timeings)
    assert any(timing['config'] == {'d0_tile': 32, 'd1_tile': 32, 'thread_num': 32} for timing in timeings)
    assert any(timing['config'] == {'d0_tile': 64, 'd1_tile': 64, 'thread_num': 64} for timing in timeings)
    assert all(timing['n'] == 10 for timing in timeings)
    assert all(timing['std'] >= 0. for timing in timeings)
    # assert all(timing['time'] > 0. for timing in timeings)

def test_tuner_allow_failure(add_one_kernel):
    X = torch.randn(1024, 1024, device="cuda")
    Y = torch.empty_like(X)
    timeings = kerneltune(
        add_one_kernel,
        {'X': X, 'Y': Y},
        configs=[{'d0_tile': 16, 'd1_tile': 16, 'thread_num': 32}, {'d0_tile': 32, 'd1_tile': 32, 'thread_num': 32}, {'d0_tile': 64, 'd1_tile': 64, 'thread_num': 19999}],
        num=10,
        no_side_effect=True,
        allow_failure=True)
    assert len(timeings) == 2
    assert any(timing['config'] == {'d0_tile': 16, 'd1_tile': 16, 'thread_num': 32} for timing in timeings)
    assert any(timing['config'] == {'d0_tile': 32, 'd1_tile': 32, 'thread_num': 32} for timing in timeings)
    assert all(timing['n'] == 10 for timing in timeings)
    assert all(timing['std'] >= 0. for timing in timeings)
    # assert all(timing['time'] > 0. for timing in timeings)

def test_tune_and_update(add_one_kernel, tmp_path):
    X = torch.randn(1024, 1024, device="cuda")
    Y = torch.empty_like(X)
    hash_fn = lambda args: str(tuple(args['X'].shape) + tuple(args['X'].stride()))
    cache = ConfigTimingCache(get_fn_name(add_one_kernel), hash_fn, root=str(tmp_path))
    configs = [{'d0_tile': 16, 'd1_tile': 16, 'thread_num': 32}, {'d0_tile': 32, 'd1_tile': 32, 'thread_num': 32}, {'d0_tile': 64, 'd1_tile': 64, 'thread_num': 64}]
    timeings = tune_and_update(
        add_one_kernel,
        {'X': X, 'Y': Y},
        configs=configs,
        cache=cache,
        num=10,
        no_side_effect=True)
    assert len(timeings) == 3
    assert any(timing['config'] == {'d0_tile': 16, 'd1_tile': 16, 'thread_num': 32} for timing in timeings)
    assert any(timing['config'] == {'d0_tile': 32, 'd1_tile': 32, 'thread_num': 32} for timing in timeings)
    assert any(timing['config'] == {'d0_tile': 64, 'd1_tile': 64, 'thread_num': 64} for timing in timeings)
    assert all(timing['n'] == 10 for timing in timeings)
    assert all(timing['std'] >= 0. for timing in timeings)
    # assert all(timing['time'] > 0. for timing in timeings)

    # check the cache
    args = {'X': X, 'Y': Y}
    assert len(cache) == 1
    assert args in cache
    assert any(cache[args] == config for config in configs)





import torch as th
from vidrial.kernels.sympow.dimensions import sympow_op_shape, sympow_dim
from vidrial.kernels.mma_sympow_bwd.op import op
from vidrial.kernels.sympow_bwd.op import op_reference as sympow_bwd_op_reference

def interface(A, B, c, expand_dim, power, d_tile, scale_A=1.0, duplicate_correction=True):
    """ First computes C_dot=A@B matmul and then c_dot = sympow_bwd(c, C_dot)
        (operation is used to implement mma_sympow_bwd)
    The kernel assumes M=D=sympow_dim(d, power, d_tile)
    Arguments:
      A: [P, M, K]
      B: [P, N, K]
      c: [P, d, N]
    Returns:
      c_dot: [P, d, N]
    """
    expand_dim = expand_dim if expand_dim < 0 else expand_dim - A.ndim
    assert expand_dim in [-2, -1], f"expand_dim={expand_dim} invalid. Only M,N dimensions can be expanded"
    transpose = expand_dim == -1 # the op only accepts expand_dim=-2. To handle -1 we need to transpose M and N
    if transpose:
        expand_dim = -2
        c = c.transpose(-1, -2)
        A, B = B.transpose(-1,-2), A.transpose(-1,-2)
    # contiguous is needed to make kernel fast
    # TODO: remove contiguous call
    c_dot = op(A.contiguous(), B.contiguous(), c.contiguous(), expand_dim, power, d_tile, scale_A, duplicate_correction)
    if transpose:
        c_dot = c_dot.transpose(-1, -2)
    return c_dot


def interface_reference(A, B, c, expand_dim, power, d_tile, scale_A=1.0, duplicate_correction=True):
    transpose = expand_dim == -2 # tiled_sympow_bwd_reference_ assumes that expand_dim=-1 (the N dim)
    if transpose: 
        expand_dim = -1
        c = c.transpose(-1, -2)
        A, B = B.transpose(-1,-2), A.transpose(-1,-2)
    C_dot = A @ B 
    C_dot = C_dot.reshape(sympow_op_shape(c.shape, power, d_tile))
    c_dot = sympow_bwd_op_reference(c, C_dot, power, d_tile, duplicate_correction)
    c_dot *= scale_A
    if transpose:
        c_dot = c_dot.transpose(-1, -2)
    return c_dot




# Query the GPU properties of the current machine
# This file is inspired by https://gist.github.com/f0k/63a664160d016a491b2cbea15913d549
import logging
import ctypes
from functools import partial
from enum import Enum
from typing import Any
logger = logging.getLogger(__name__)

CUDA_SUCCESS = 0

# Constants taken from cuda.h
class CUDA_DEVICE_ATTRIBUTE(Enum):
    CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_BLOCK = 1                          # Maximum number of threads per block 
    CU_DEVICE_ATTRIBUTE_MAX_BLOCK_DIM_X = 2                                # Maximum block dimension X 
    CU_DEVICE_ATTRIBUTE_MAX_BLOCK_DIM_Y = 3                                # Maximum block dimension Y 
    CU_DEVICE_ATTRIBUTE_MAX_BLOCK_DIM_Z = 4                                # Maximum block dimension Z 
    CU_DEVICE_ATTRIBUTE_MAX_GRID_DIM_X = 5                                 # Maximum grid dimension X 
    CU_DEVICE_ATTRIBUTE_MAX_GRID_DIM_Y = 6                                 # Maximum grid dimension Y 
    CU_DEVICE_ATTRIBUTE_MAX_GRID_DIM_Z = 7                                 # Maximum grid dimension Z 
    CU_DEVICE_ATTRIBUTE_MAX_SHARED_MEMORY_PER_BLOCK = 8                    # Maximum shared memory available per block in bytes 
    CU_DEVICE_ATTRIBUTE_TOTAL_CONSTANT_MEMORY = 9                          # Memory available on device for __constant__ variables in a CUDA C kernel in bytes 
    CU_DEVICE_ATTRIBUTE_WARP_SIZE = 10                                     # Warp size in threads 
    CU_DEVICE_ATTRIBUTE_MAX_PITCH = 11                                     # Maximum pitch in bytes allowed by memory copies 
    CU_DEVICE_ATTRIBUTE_MAX_REGISTERS_PER_BLOCK = 12                       # Maximum number of 32-bit registers available per block 
    CU_DEVICE_ATTRIBUTE_CLOCK_RATE = 13                                    # Typical clock frequency in kilohertz 
    CU_DEVICE_ATTRIBUTE_TEXTURE_ALIGNMENT = 14                             # Alignment requirement for textures 
    CU_DEVICE_ATTRIBUTE_MULTIPROCESSOR_COUNT = 16                          # Number of multiprocessors on device 
    CU_DEVICE_ATTRIBUTE_KERNEL_EXEC_TIMEOUT = 17                           # Specifies whether there is a run time limit on kernels 
    CU_DEVICE_ATTRIBUTE_INTEGRATED = 18                                    # Device is integrated with host memory 
    CU_DEVICE_ATTRIBUTE_CAN_MAP_HOST_MEMORY = 19                           # Device can map host memory into CUDA address space 
    CU_DEVICE_ATTRIBUTE_COMPUTE_MODE = 20                                  # Compute mode (See ::CUcomputemode for details) 
    CU_DEVICE_ATTRIBUTE_CONCURRENT_KERNELS = 31                            # Device can possibly execute multiple kernels concurrently 
    CU_DEVICE_ATTRIBUTE_ECC_ENABLED = 32                                   # Device has ECC support enabled 
    CU_DEVICE_ATTRIBUTE_MEMORY_CLOCK_RATE = 36                             # Peak memory clock frequency in kilohertz 
    CU_DEVICE_ATTRIBUTE_GLOBAL_MEMORY_BUS_WIDTH = 37                       # Global memory bus width in bits 
    CU_DEVICE_ATTRIBUTE_L2_CACHE_SIZE = 38                                 # Size of L2 cache in bytes 
    CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_MULTIPROCESSOR = 39                # Maximum resident threads per multiprocessor 
    CU_DEVICE_ATTRIBUTE_UNIFIED_ADDRESSING = 41                            # Device shares a unified address space with the host 
    CU_DEVICE_ATTRIBUTE_COMPUTE_CAPABILITY_MAJOR = 75                      # Major compute capability version number 
    CU_DEVICE_ATTRIBUTE_COMPUTE_CAPABILITY_MINOR = 76                      # Minor compute capability version number 
    CU_DEVICE_ATTRIBUTE_STREAM_PRIORITIES_SUPPORTED = 78                   # Device supports stream priorities 
    CU_DEVICE_ATTRIBUTE_GLOBAL_L1_CACHE_SUPPORTED = 79                     # Device supports caching globals in L1 
    CU_DEVICE_ATTRIBUTE_LOCAL_L1_CACHE_SUPPORTED = 80                      # Device supports caching locals in L1 
    CU_DEVICE_ATTRIBUTE_MAX_SHARED_MEMORY_PER_MULTIPROCESSOR = 81          # Maximum shared memory available per multiprocessor in bytes 
    CU_DEVICE_ATTRIBUTE_MAX_REGISTERS_PER_MULTIPROCESSOR = 82              # Maximum number of 32-bit registers available per multiprocessor 
    CU_DEVICE_ATTRIBUTE_MANAGED_MEMORY = 83                                # Device can allocate managed memory on this system 
    CU_DEVICE_ATTRIBUTE_MULTI_GPU_BOARD = 84                               # Device is on a multi-GPU board 
    CU_DEVICE_ATTRIBUTE_MULTI_GPU_BOARD_GROUP_ID = 85                      # Unique id for a group of devices on the same multi-GPU board 
    CU_DEVICE_ATTRIBUTE_SINGLE_TO_DOUBLE_PRECISION_PERF_RATIO = 87         # Ratio of single precision performance (in floating-point operations per second) to double precision performance 
    CU_DEVICE_ATTRIBUTE_CONCURRENT_MANAGED_ACCESS = 89                     # Device can coherently access managed memory concurrently with the CPU 
    CU_DEVICE_ATTRIBUTE_CAN_USE_HOST_POINTER_FOR_REGISTERED_MEM = 91       # Device can access host registered memory at the same virtual address as the CPU 
    CU_DEVICE_ATTRIBUTE_COOPERATIVE_LAUNCH = 95                            # Device supports launching cooperative kernels via ::cuLaunchCooperativeKernel 
    CU_DEVICE_ATTRIBUTE_MAX_SHARED_MEMORY_PER_BLOCK_OPTIN = 97             # Maximum optin shared memory per block 
    CU_DEVICE_ATTRIBUTE_MAX_BLOCKS_PER_MULTIPROCESSOR = 106                # Maximum number of blocks per multiprocessor 
    CU_DEVICE_ATTRIBUTE_MAX_PERSISTING_L2_CACHE_SIZE = 108                 # Maximum L2 persisting lines capacity setting in bytes. 
    CU_DEVICE_ATTRIBUTE_RESERVED_SHARED_MEMORY_PER_BLOCK = 111             # Shared memory reserved by CUDA driver per block in bytes 
    CU_DEVICE_ATTRIBUTE_TENSOR_MAP_ACCESS_SUPPORTED = 127                  # Device supports accessing memory using Tensor Map. 


def ConvertSMVer2Cores(major, minor):
    # Returns the number of CUDA cores per multiprocessor for a given
    # Compute Capability version. There is no way to retrieve that via
    # the API, so it needs to be hard-coded.
    # See _ConvertSMVer2Cores in helper_cuda.h in NVIDIA's CUDA Samples.
    return {(1, 0): 8,    # Tesla
            (1, 1): 8,
            (1, 2): 8,
            (1, 3): 8,
            (2, 0): 32,   # Fermi
            (2, 1): 48,
            (3, 0): 192,  # Kepler
            (3, 2): 192,
            (3, 5): 192,
            (3, 7): 192,
            (5, 0): 128,  # Maxwell
            (5, 2): 128,
            (5, 3): 128,
            (6, 0): 64,   # Pascal
            (6, 1): 128,
            (6, 2): 128,
            (7, 0): 64,   # Volta
            (7, 2): 64,
            (7, 5): 64,   # Turing
            (8, 0): 64,   # Ampere
            (8, 6): 128,
            (8, 7): 128,
            (8, 9): 128,  # Ada
            (9, 0): 128,  # Hopper
            }.get((major, minor), 0)


def get_cuda_device_properties(attrs) -> list[dict[str, Any]]:
    libnames = ('libcuda.so', 'libcuda.dylib', 'nvcuda.dll', 'cuda.dll')
    for libname in libnames:
        try:
            cuda = ctypes.CDLL(libname)
        except OSError:
            continue
        else:
            break
    else:
        raise OSError("could not load any of: " + ' '.join(libnames))

    nGpus = ctypes.c_int()
    name = b' ' * 100
    cc_major = ctypes.c_int()
    cc_minor = ctypes.c_int()
    threads_per_core = ctypes.c_int()
    freeMem = ctypes.c_size_t()
    totalMem = ctypes.c_size_t()

    result = ctypes.c_int()
    device = ctypes.c_int()
    context = ctypes.c_void_p()
    error_str = ctypes.c_char_p()

    result = cuda.cuInit(0)
    if result != CUDA_SUCCESS:
        cuda.cuGetErrorString(result, ctypes.byref(error_str))
        logger.error("cuInit failed with error code %d: %s" % (result, error_str.value.decode()))
        return 1
    result = cuda.cuDeviceGetCount(ctypes.byref(nGpus))
    if result != CUDA_SUCCESS:
        cuda.cuGetErrorString(result, ctypes.byref(error_str))
        logger.error("cuDeviceGetCount failed with error code %d: %s" % (result, error_str.value.decode()))
        return 1
    logger.info("Found %d device(s)." % nGpus.value)
    res = [{} for _ in range(int(nGpus.value))]
    for i in range(nGpus.value):
        result = cuda.cuDeviceGet(ctypes.byref(device), i)
        if result != CUDA_SUCCESS:
            cuda.cuGetErrorString(result, ctypes.byref(error_str))
            logger.error("cuDeviceGet failed with error code %d: %s" % (result, error_str.value.decode()))
            return 1
        logger.info("Device: %d" % i)
        if cuda.cuDeviceGetName(ctypes.c_char_p(name), len(name), device) == CUDA_SUCCESS:
            res[i]['name'] = name.split(b'\0', 1)[0].decode()
        if cuda.cuDeviceComputeCapability(ctypes.byref(cc_major), ctypes.byref(cc_minor), device) == CUDA_SUCCESS:
            res[i]['compute_capability'] = (cc_major.value, cc_minor.value)
        for attr in attrs:
            v = ctypes.c_int()
            if cuda.cuDeviceGetAttribute(ctypes.byref(v), attr.value, device) == CUDA_SUCCESS:    
                res[i][attr.name] = v.value
                if attr == CUDA_DEVICE_ATTRIBUTE.CU_DEVICE_ATTRIBUTE_MULTIPROCESSOR_COUNT:
                    res[i]['cuda_cores'] = v.value * ConvertSMVer2Cores(cc_major.value, cc_minor.value)
                    if cuda.cuDeviceGetAttribute(ctypes.byref(threads_per_core), CUDA_DEVICE_ATTRIBUTE.CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_MULTIPROCESSOR.value, device) == CUDA_SUCCESS:
                        res[i]['cuda_threads'] = v.value * threads_per_core.value
                if attr == CUDA_DEVICE_ATTRIBUTE.CU_DEVICE_ATTRIBUTE_CLOCK_RATE:
                    res[i]['clock_rate (MHz)'] = v.value / 1000.
                if attr == CUDA_DEVICE_ATTRIBUTE.CU_DEVICE_ATTRIBUTE_MEMORY_CLOCK_RATE:
                    res[i]['memory_clock_rate (MHz)'] = v.value / 1000.


        try:
            result = cuda.cuCtxCreate_v2(ctypes.byref(context), 0, device)
        except AttributeError:
            result = cuda.cuCtxCreate(ctypes.byref(context), 0, device)
        if result != CUDA_SUCCESS:
            cuda.cuGetErrorString(result, ctypes.byref(error_str))
            logger.error("cuCtxCreate failed with error code %d: %s" % (result, error_str.value.decode()))
        else:
            try:
                result = cuda.cuMemGetInfo_v2(ctypes.byref(freeMem), ctypes.byref(totalMem))
            except AttributeError:
                result = cuda.cuMemGetInfo(ctypes.byref(freeMem), ctypes.byref(totalMem))
            if result == CUDA_SUCCESS:
                res[i]['total_memory (MiB)'] = totalMem.value / 1024**2
                res[i]['free_memory (MiB)'] = freeMem.value / 1024**2
            else:
                cuda.cuGetErrorString(result, ctypes.byref(error_str))
                logger.error("cuMemGetInfo failed with error code %d: %s" % (result, error_str.value.decode()))
            cuda.cuCtxDetach(context)
    return res


get_cuda_device_basic_props = partial(get_cuda_device_properties, [
    CUDA_DEVICE_ATTRIBUTE.CU_DEVICE_ATTRIBUTE_MULTIPROCESSOR_COUNT,
    CUDA_DEVICE_ATTRIBUTE.CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_MULTIPROCESSOR,
    CUDA_DEVICE_ATTRIBUTE.CU_DEVICE_ATTRIBUTE_CLOCK_RATE,
    CUDA_DEVICE_ATTRIBUTE.CU_DEVICE_ATTRIBUTE_MEMORY_CLOCK_RATE,
    CUDA_DEVICE_ATTRIBUTE.CU_DEVICE_ATTRIBUTE_MAX_SHARED_MEMORY_PER_MULTIPROCESSOR,
    CUDA_DEVICE_ATTRIBUTE.CU_DEVICE_ATTRIBUTE_MAX_REGISTERS_PER_MULTIPROCESSOR,
])



import torch as th
from vidrial.kernels.sympow.interface import interface_reference as sympow_reference
from vidrial.kernels.sympow.dimensions import sympow_shape
from vidrial.kernels.mma_sympow_bwd.interface import interface as mma_sympow_bwd
from vidrial.kernels.sympow_mma.binding import binding
from vidrial.kernels.sympow_mma.dimensions import op_output_shape

@th.library.custom_op("vidrial::sympow_mma", mutates_args=())
def op(A: th.Tensor, B: th.Tensor, expand_dim: int, power: int, d_tile: int = 1, scale_A: float = 1.0, duplicate_correction: bool = True) -> th.Tensor:
    """ Compute sympow(A,power,d_tile,expand_dim) @ B

    Input shapes:
        A: [P, M, d] or [P, d, K] - Left matrix
        B: [P, K, N] - Right matrix
        expand_dim: int - Dimension to expand in A (-1 for expand K, -2 for expand M)
        power: int - Power of the symmetric power
        d_tile: int - Tile size for the operation
        duplicate_correction: bool - Whether to correct for duplicate elements

    Output shape:
        C: [P, M, N] - Result of expanded A @ B
    """
    C_shape = op_output_shape(A.shape, B.shape, expand_dim, power, d_tile)
    C = th.empty(C_shape, device=A.device, dtype=A.dtype)
    binding(A, B, C, expand_dim, power, d_tile, scale_A, duplicate_correction) # type: ignore
    return C

@op.register_fake
def op_fake(A: th.Tensor, B: th.Tensor, expand_dim: int, power: int, d_tile: int, scale_A: float = 1.0, duplicate_correction: bool = True):
    eA_shape = sympow_shape(A.shape, power, d_tile, expand_dim%A.ndim)
    return th.empty((*A.shape[:-2], eA_shape[-2], B.shape[-1]), device=A.device, dtype=A.dtype)


# ------------- Backward Implementation -------------

def op_setup(ctx, inputs, output):
    A, B, expand_dim, power, d_tile, scale_A, duplicate_correction = inputs
    ctx.save_for_backward(A, B)
    ctx.expand_dim = expand_dim
    ctx.power = power
    ctx.d_tile = d_tile
    ctx.scale_A = scale_A
    ctx.duplicate_correction = duplicate_correction
def bwd_dA(A, B, dC, expand_dim, power, d_tile, scale_A, duplicate_correction):
    A, B, c = dC, B.transpose(-1, -2), A
    return mma_sympow_bwd(A, B, c, expand_dim, power, d_tile, scale_A, duplicate_correction)
def bwd_dB(A, B, dC, expand_dim, power, d_tile, scale_A, duplicate_correction):
    expand_dim = {-1: -2, -2: -1}[expand_dim%A.ndim-A.ndim]
    A = A.transpose(-1, -2)
    return op(A, dC, expand_dim, power, d_tile, scale_A, duplicate_correction)
def op_backward(ctx, dC):
    A, B = ctx.saved_tensors
    dA = bwd_dA(A, B, dC, ctx.expand_dim, ctx.power, ctx.d_tile, ctx.scale_A, ctx.duplicate_correction)
    dB = bwd_dB(A, B, dC, ctx.expand_dim, ctx.power, ctx.d_tile, ctx.scale_A, ctx.duplicate_correction)
    assert dA.shape == A.shape
    assert dB.shape == B.shape
    return dA, dB, None, None, None, None
th.library.register_autograd("vidrial::sympow_mma", op_backward, setup_context=op_setup)


# ------------- Reference Implementation -------------

def op_reference(A: th.Tensor, B: th.Tensor, dim, power, d_tile=1, scale_A=1.0, duplicate_correction=True):
    eA = sympow_reference(A, power, d_tile, dim, duplicate_correction)
    eA = eA * scale_A
    return eA @ B




from vidrial.kernels.sympow_mma.op import op, op_reference

""" Functional interface """
def sympowA_mm(A, B, dim, power, d_tile):
    return op(A, B, dim, power, d_tile)
def sympowA_mma(A, B, C, dim, power, d_tile):
    return C + sympowA_mm(A, B, dim, power, d_tile)
def sympowA_gemm(α, A, B, β, C, dim, power, d_tile):
    return β * C + α * sympowA_mm(A, B, dim, power, d_tile)

def sympowA_mm_reference(A, B, dim, power, d_tile):
    return op_reference(A, B, dim, power, d_tile)
def sympowA_mma_reference(A, B, C, dim, power, d_tile):
    """ A, B, C -> D = C + A @ B """
    return C + sympowA_mm_reference(A, B, dim, power, d_tile)
def sympowA_gemm_reference(α, A, B, β, C, dim, power, d_tile):
    """ α, A, B, β, C -> D = β * C + α * A @ B """
    return β * C + α * sympowA_mm_reference(A, B, dim, power, d_tile)


""" Inplace interface """
# def sympowA_mm_(A, B, D, dim, power, d_tile):
#     D[:] = sympowA_mm(A, B, dim, power, d_tile)
def sympowA_mma_(A, B, C, dim, power, d_tile):
    C[:] = sympowA_mma(A, B, C, dim, power, d_tile)
def sympowA_gemm_(α, A, B, β, C, dim, power, d_tile):
    C[:] = sympowA_gemm(α, A, B, β, C, dim, power, d_tile)

def sympowA_mm_reference_(A, B, C, dim, power, d_tile):
    C[:] = sympowA_mm_reference(A, B, dim, power, d_tile)
def sympowA_mma_reference_(A, B, C, dim, power, d_tile):
    C[:] = sympowA_mma_reference(A, B, C, dim, power, d_tile)
def sympowA_gemm_reference_(α, A, B, β, C, dim, power, d_tile):
    C[:] = sympowA_gemm_reference(α, A, B, β, C, dim, power, d_tile)



import pytest
import torch as th
from vidrial.kernels.sympow.dimensions import sympow_dim
from vidrial.py_utils.test_utils import diff, PERMUTATIONS_OF_STRIDES_UP_TO_4D, create_permuted_strides_layout
from vidrial.kernels.sympow.interface import interface as sympow
from vidrial.kernels.sympow.interface import interface_reference as sympow_reference
from vidrial.kernels.sympow_mma.binding import binding
from vidrial.kernels.sympow_mma.op import op, op_reference

import logging

logging.basicConfig(level=logging.DEBUG) 


PROBLEM_SHAPES = [
    ((3,2), 16, 64, 32, 2, 8, False),
    ((1,), 16, 64, 32, 2, 8, False),
    ((1,1), 16, 64, 32, 2, 8, True),
    ((10,), 16, 1, 32, 2, 8, True),
    ((2,), 16, 64, 1, 2, 8, True),
    ((2,3,4), 32, 10, 4, 3, 4, True),
    ((6,), 64, 1, 128, 2, 8, True),
]

def setup(batch_dims, d, N, R, p, d_tile, expand_dim, duplicate_correction):
    D = sympow_dim(d, p, d_tile)
    M, K = {-2: (D, R), -1: (R, D)}[expand_dim]
    A_shape = batch_dims+{-2:(d, K), -1:(M, d)}[expand_dim]
    A = th.randn(A_shape, device="cuda")
    B = th.randn(batch_dims+(K, N), device="cuda")
    C = th.randn(batch_dims+(M, N), device="cuda")
    return A, B, C

@pytest.mark.parametrize("batch_dims,d,N,R,p,d_tile,duplicate_correction", PROBLEM_SHAPES)
@pytest.mark.parametrize("expand_dim", [-1, -2 ])
def test_binding_autotuned(batch_dims, d, N, R, p, d_tile, expand_dim, duplicate_correction):
    A, B, C = setup(batch_dims, d, N, R, p, d_tile, expand_dim, duplicate_correction)
    binding(A, B, C, expand_dim, p, d_tile, 1., duplicate_correction) # type: ignore
    eA_ref = sympow_reference(A, p, d_tile, expand_dim, duplicate_correction)
    C_ref = eA_ref @ B
    diff(C, C_ref, atol=1e-1, rtol=1e-1)

@pytest.mark.parametrize("batch_dims,d,N,R,p,d_tile,duplicate_correction", PROBLEM_SHAPES)
@pytest.mark.parametrize("expand_dim", [-1, -2 ])
def test_op(batch_dims, d, N, R, p, d_tile, expand_dim, duplicate_correction):
    A, B, dC = setup(batch_dims, d, N, R, p, d_tile, expand_dim, duplicate_correction)
    A.requires_grad, B.requires_grad = True, True
    C = op(A, B, expand_dim, p, d_tile, duplicate_correction)
    C.backward(dC)
    assert A.grad is not None and B.grad is not None
    dA, dB = A.grad.clone(), B.grad.clone()
    A.grad, B.grad = None, None

    C_ref = op_reference(A, B, expand_dim, p, d_tile, duplicate_correction)
    C_ref.backward(dC)
    assert A.grad is not None and B.grad is not None
    dA_ref, dB_ref = A.grad.clone(), B.grad.clone()
    A.grad, B.grad = None, None

    diff(C, C_ref, atol=1e-1, rtol=1e-1)
    diff(dA, dA_ref, atol=1e-1, rtol=1e-1)
    diff(dB, dB_ref, atol=1e-1, rtol=1e-1)

def test_non_standard_strides(batch_dims=(2,3,4), d=16, N=16, R=32, p=2, d_tile=16, expand_dim=-1, duplicate_correction=True):
    A, B, C = setup(batch_dims, d, N, R, p, d_tile, expand_dim, duplicate_correction)
    non_standard_stride = lambda X: X.transpose(2, 1).contiguous().transpose(2, 1)
    A, B, C = non_standard_stride(A), non_standard_stride(B), non_standard_stride(C)
    binding(A, B, C, expand_dim, p, d_tile, 1., duplicate_correction) # type: ignore
    eA_ref = sympow_reference(A, p, d_tile, expand_dim, duplicate_correction)
    C_ref = eA_ref @ B
    diff(C, C_ref, atol=1e-1, rtol=1e-1)



# import torch as th
# from tyle.kernels.sympow.reference import *

# print("\nThe tpow and sympow functions");
# x = th.arange(3)
# y = tpow(x, 2)
# z = sympow(x, 2, duplicate_correction=False)
# print("x: ", x)
# print("y: ", y)
# print("z: ", z)
# print()
# x = th.arange(4)
# y = tpow(x, 2)
# z = sympow(x, 2, duplicate_correction=False)
# print("x: ", x)
# print("y: ", y)
# print("z: ", z)

# print("\nHigher powers");
# x = th.arange(4)
# y = tpow(x, 3)
# z = sympow(x, 3, duplicate_correction=False)
# print("x.shape: ", x.shape)
# print("y.shape: ", y.shape)
# print("z.shape: ", z.shape)

# print("\nWith batch dimension");
# x = th.arange(12*4).reshape(12,4) % 8
# y = tpow(x, 2)
# z = sympow(x, 2, duplicate_correction=False)
# print('x.shape: ', x.shape)
# print("y.shape: ", y.shape)
# print("z.shape: ", z.shape)

# print("\nHigh rank symmetric tensors:")
# print(y.sum(0))
# print(z.sum(0))

# print("\nArbitrary number of batch dimensions");
# x = th.randn(2,3,4)
# y = tpow(x, 2)
# z = sympow(x, 2, duplicate_correction=False)
# print('x.shape: ', x.shape)
# print("y.shape: ", y.shape)
# print("z.shape: ", z.shape)

# print("\nPreserving the norm of tpow instead of the elements");
# x = th.rand(4)
# print("norm(y): ", th.linalg.norm(tpow(x, 2)))
# print("With duplicate_correction=False:")
# print("norm(z): ", th.linalg.norm(sympow(x, 2, duplicate_correction=False)))
# print("With duplicate_correction=True:")
# print("norm(z): ", th.linalg.norm(sympow(x, 2, duplicate_correction=True)))

# print("\nRunning on the GPU:")
# x = th.randn(4, 8, device="cuda")
# z_gpu = sympow(x, power=2)
# z_cpu = sympow(x.cpu().clone(), power=2)
# print("allclose(z_gpu, z_cpu): ", th.allclose(z_gpu.cpu(), z_cpu))
# x = th.randn(4, 16, device="cuda")
# z_gpu = sympow(x, power=3)
# z_cpu = sympow(x.cpu().clone(), power=3)
# print("allclose(z_gpu, z_cpu): ", th.allclose(z_gpu.cpu(), z_cpu))


# print("\nTiled sympow: interpolating between sympow and tpow")
# x = th.randn(2,16)
# z = sympow(x, power=2)
# z_tile1 = sympow(x, power=2, d_tile=1)
# z_tile2 = sympow(x, power=2, d_tile=2)
# z_tile4 = sympow(x, power=2, d_tile=4)
# z_tile8 = sympow(x, power=2, d_tile=8)
# z_tile16 = sympow(x, power=2, d_tile=16)
# y = tpow(x, power=2)
# print("x.shape: ", list(x.shape))
# print("z.shape: ", list(z.shape))
# print("z_tile1.shape: ", list(z_tile1.shape))
# print("z_tile2.shape: ", list(z_tile2.shape))
# print("z_tile4.shape: ", list(z_tile4.shape))
# print("z_tile8.shape: ", list(z_tile8.shape))
# print("z_tile16.shape: ", list(z_tile16.shape))
# print("y.shape: ", list(y.shape))

# print("\nsympow_expand arbitary rank tensors")
# x = th.randn(2,3,4,5)
# print("x.shape: ", list(x.shape))
# for i in range(len(x.shape)):
#     ex = sympow_expand(x, expand_dim=i, power=2)
#     print(f"expand_dim={i}, ex.shape={list(ex.shape)}")

# print("\nsympow_expand works with tiled sympow")
# x = th.randn(2,4,8,16)
# print("x.shape: ", list(x.shape))
# for i in range(len(x.shape)):
#     ex = sympow_expand(x, expand_dim=i, power=2)
#     ex_tile_2 = sympow_expand(x, expand_dim=i, power=2, d_tile=2)
#     print(f"expand_dim={i}, ex.shape={list(ex.shape)}, ex_tile_2.shape={list(ex_tile_2.shape)}")

# print("\nFused sympow mma: expanding the -2 dimension of A")
# A = th.randn(2, 16, 16, device="cuda")
# B = th.randn(2, 16, 32, device="cuda")
# eA = sympow_expand(A.cpu(), -2, power=2, d_tile=4)
# C_ref = eA @ B.cpu()
# fused_C = sympow_expandA_mma(A, B, -2, power=2, d_tile=4)
# print("A.shape: ", list(A.shape))
# print("eA.shape: ", list(eA.shape))
# print("B.shape: ", list(B.shape))
# print("C.shape: ", list(C_ref.shape))
# tol = 1e-2
# print("allclose(fused_C, C_ref): ", th.allclose(fused_C.cpu(), C_ref, atol=tol, rtol=tol))
 
# print("\nFused sympow mma: expanding the -1 dimension of A")
# A = th.randn(2, 16, 16, device="cuda")
# B = th.randn(2, 160, 32, device="cuda")
# eA = sympow_expand(A.cpu(), -1, power=2, d_tile=4)
# C_ref = eA @ B.cpu()
# fused_C = sympow_expandA_mma(A, B, -1, power=2, d_tile=4)
# print("A.shape: ", list(A.shape))
# print("eA.shape: ", list(eA.shape))
# print("B.shape: ", list(B.shape))
# print("C.shape: ", list(C_ref.shape))
# tol = 1e-1
# print("allclose(fused_C, C_ref): ", th.allclose(fused_C.cpu(), C_ref, atol=tol, rtol=tol))
# print("fused_C: ", fused_C.flatten()[:500:10])
# print("C_ref: ", C_ref.flatten()[:500:10])
 


import pytest
import torch as th
from math import prod
from vidrial.kernels.reduce.binding import binding
from vidrial.py_utils.test_utils import diff

class SumReduce:
    str = "return a + b;"
    reference = lambda X, reduce_dims: th.sum(X, dim=reduce_dims)

class MaxReduce:
    str = "return a > b ? a : b;"
    @staticmethod
    def reference(X, reduce_dims):
        for i in reversed(sorted(reduce_dims)):
            X = th.max(X, dim=i).values
        return X

@pytest.mark.parametrize(
    ["dtype", "reduce_fn", "X_shape", "reduce_dims", "X_tile_shape", "thread_num"],
    [
        (int, SumReduce, (4, 8), (0,), None, 1),
        (int, SumReduce, (4, 8), (0,), (1,1), 1),
        (int, SumReduce, (4, 8), (1,), None, 2),
        (int, SumReduce, (4, 8, 16), (2,), None, 4),
        # (int, SumReduce, (4, 8), (0, 1), None, 8), # TODO: support reducing all the dimensions
        (th.int32, SumReduce, (4, 8, 16), (0,2), None, 16),
        (th.int32, SumReduce, (128, 32), (0,), None, 64),
        (th.int32, SumReduce, (128, 32), (0,), None, 32),
        (th.int32, SumReduce, (128, 32), (0,), (32, 32), 32),
        (th.int32, SumReduce, (16, 8), (0,), None, 32),
        (th.int32, SumReduce, (4, 8), (0,), None, 32),
        (th.int32, SumReduce, (128, 1024), (0,), (128, 32), 32),
        (th.int32, SumReduce, (128, 32), (0,), (128, 32), 32),
        (th.int32, SumReduce, (128, 64), (0,), (128, 32), 32),
        (th.float, SumReduce, (256, 1024), (1,), (4, 1024), 32),
        (th.float, SumReduce, (256, 1024), (0,), (4, 1024), 32),
        (th.float16, SumReduce, (128, 1024), (0,), (128, 32), 32),
        (int, MaxReduce, (4, 8), (0,), None, 1),
        (int, MaxReduce, (4, 8), (1,), None, 2),
        (int, MaxReduce, (4, 8, 16), (2,), None, 4),
        (th.uint8, MaxReduce, (4, 8, 16), (0,1), None, 16),
        (th.int32, MaxReduce, (128, 32), (0,), None, 64),
        (th.int32, MaxReduce, (128, 32), (0,), (64, 32), 64),
        (th.int32, MaxReduce, (128, 32), (0,), (16, 16), 32),
        (th.int32, MaxReduce, (16, 8), (0,), None, 32),
    ]
)
def test_binding(dtype, reduce_fn, X_shape, reduce_dims, X_tile_shape, thread_num):
    X_tile_shape = X_shape if X_tile_shape is None else X_tile_shape
    x_shape = [X_shape[i] for i in range(len(X_shape)) if i not in reduce_dims]
    X = th.randn(X_shape, device='cuda').to(dtype)
    x = th.empty(x_shape, device='cuda', dtype=dtype)
    binding(X, x, reduce_fn.str, reduce_dims, X_tile_shape, thread_num)
    x_ref = reduce_fn.reference(X, reduce_dims)
    tol = 1e-1 if dtype == th.float16 else 1e-4
    diff(x, x_ref, atol=tol, rtol=tol)


from typing import Union, Any

class BaseType:
    """Base class for all C++ type representations."""
    def to_cpp(self) -> str:
        """Generate C++ string representation of the type."""
        raise NotImplementedError


    def __repr__(self) -> str:
        return str(self)

class Int(BaseType):
    """Represents a compile-time integer value."""
    def __init__(self, value: int):
        self.value = value

    def to_cpp(self, _=None) -> str:
        return f"Int<{self.value}>"
    
    def size(self) -> int:
        return self.value
    
    def __mul__(self, other: Union[int, 'Int']) -> 'Int':
        if isinstance(other, Int):
            return Int(self.value * other.value)
        else:
            return Int(self.value * other)
        
    def __rmul__(self, other: Union[int, 'Int']) -> 'Int':
        return self * other
    
    def __int__(self) -> int:
        return self.value
    
    def __str__(self) -> str:
        return str(self.value)
    
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(value={self.value})"
    
    def __reduce__(self):
        """Make Int picklable by specifying how to reconstruct it."""
        return (type(self), (self.value,))
    
    def __eq__(self, other: Union[int, 'Int']) -> bool:
        if isinstance(other, Int):
            return self.value == other.value
        elif isinstance(other, int):
            return self.value == other
        else:
            raise ValueError(f"Cannot compare Int with {type(other)}")
        
    def __le__(self, other: Union[int, 'Int']) -> bool:
        if isinstance(other, Int):
            return self.value <= other.value
        elif isinstance(other, int):
            return self.value <= other
        else:
            raise ValueError(f"Cannot compare Int with {type(other)}")
        
    def __lt__(self, other: Union[int, 'Int']) -> bool:
        if isinstance(other, Int):
            return self.value < other.value
        elif isinstance(other, int):
            return self.value < other
        else:
            raise ValueError(f"Cannot compare Int with {type(other)}")
        
    def __ge__(self, other: Union[int, 'Int']) -> bool:
        if isinstance(other, Int):
            return self.value >= other.value
        elif isinstance(other, int):
            return self.value >= other
        else:
            raise ValueError(f"Cannot compare Int with {type(other)}")
        
    def __gt__(self, other: Union[int, 'Int']) -> bool:
        if isinstance(other, Int):
            return self.value > other.value
        elif isinstance(other, int):
            return self.value > other
        else:
            raise ValueError(f"Cannot compare Int with {type(other)}")

    def __pow__(self, other: Union[int, 'Int']) -> Union['Int', int]:
        if isinstance(other, Int):
            return Int(self.value ** other.value)
        else:
            return int(self.value ** other)

    def __truediv__(self, other: Union[int, 'Int']) -> Union['Int', int]:
        if isinstance(other, Int):
            return Int(self.value // other.value)
        else:
            return int(self.value // other)
        
    def __floordiv__(self, other: Union[int, 'Int']) -> Union['Int', int]:
        if isinstance(other, Int):
            return Int(self.value // other.value)
        else:
            return int(self.value // other)
        
    def __mod__(self, other: Union[int, 'Int']) -> Union['Int', int]:
        if isinstance(other, Int):
            return Int(self.value % other.value)
        else:
            return int(self.value % other)
        
    def __hash__(self) -> int:
        return hash(self.value)
        
        
def intify(value: Union[int, Any]) -> Any:
    if isinstance(value, int):
        return Int(value)
    return value

_1 = Int(1)
_2 = Int(2)
_3 = Int(3)
_4 = Int(4)
_5 = Int(5)
_6 = Int(6)
_7 = Int(7)
_8 = Int(8)
_9 = Int(9)
_10 = Int(10)
_16 = Int(16)
_32 = Int(32)
_64 = Int(64)
_128 = Int(128)
_256 = Int(256)
_512 = Int(512)
_1024 = Int(1024)
_2048 = Int(2048)
_4096 = Int(4096)
_8192 = Int(8192)
_16384 = Int(16384)


import torch
from typing import Any, Optional, Tuple
from vidrial.jit.static.types import Layout, Shape, Stride, Int, Tuple
from vidrial.jit.static.types import CudaDType

def nested_constructor(nested_dims):
    def f(seq, type, dim_spec):
        if isinstance(dim_spec, int):
            # Handle negative indices
            actual_dim = dim_spec if dim_spec >= 0 else len(seq) + dim_spec
            return type(Int(seq[actual_dim]))
        elif isinstance(dim_spec, (list, tuple)):
            # For nested dimensions, recursively build shape and stride for each
            return type(*[f(seq, type, nested_dim) for nested_dim in dim_spec])
        raise ValueError(f"Unsupported dimension specification: {dim_spec}")
    return lambda seq, type=Tuple: f(seq, type, nested_dims)

def layout_from_shape_stride(shape: tuple[int, ...], stride: tuple[int, ...], dims: Optional[tuple[Any, ...]] = None) -> Layout:
    dims = tuple(range(len(shape))) if dims is None else dims
    constructor = nested_constructor(dims)
    return Layout(constructor(shape, Shape), constructor(stride, Stride))

# Convinient layout constructors
def layout_from_tensor(t: torch.Tensor, dims: Optional[tuple[Any, ...]] = None) -> Layout:
    """ Create a layout from a tensor by specifying the layout from the dimensions of the 
    tensor. For example, for a tensor of shape (b, t, h, d), if we want to create a layout 
    with shape (d, (b, (t, h))), we can pass dims=(-1, (0, (1, 2))). The strides will be
    inferred correctly.

    Args:
        t: The tensor to create the layout from.
        dims: An ordered hierarchical tuple of dimensions to use for the layout.
    
    Returns:
        A Layout object with the specified shape and stride.
    """
    constructor = nested_constructor(dims)
    return Layout(constructor(t.shape, Shape), constructor(t.stride(), Stride))


def torch_dtype_to_c(dtype: torch.dtype) -> str:
    """Convert PyTorch dtype to C++ type string."""
    dtype_map = {
        torch.bfloat16: CudaDType.BFLOAT16.value,
        torch.float16: CudaDType.FLOAT16.value,
        torch.float32: CudaDType.FLOAT.value,
        torch.float64: CudaDType.DOUBLE.value,
        torch.int32: CudaDType.INT32.value,
        torch.int64: CudaDType.INT64.value,
        torch.uint8: CudaDType.UINT8.value,
    }
    if dtype not in dtype_map:
        raise ValueError(f"Unsupported dtype: {dtype}")
    return dtype_map[dtype]

NestedStructure = Tuple['NestedStructure', ...] | int

def test_layout_from_tensor():
    t = torch.randn(1, 2, 3, 4)
    layout = layout_from_tensor(t, (-1, (0, (1, 2))))
    assert layout.to_cpp() == "Layout<Shape<Int<4>, Shape<Int<1>, Shape<Int<2>, Int<3>>>>, Stride<Int<1>, Stride<Int<24>, Stride<Int<12>, Int<4>>>>>"



import torch as th
from vidrial.kernels.mma_sympow_bwd.binding import binding 

@th.library.custom_op("vidrial::mma_sympow_bwd", mutates_args=())
def op(A: th.Tensor, B: th.Tensor, c: th.Tensor, expand_dim: int, power: int, d_tile: int, scale_A: float = 1.0, duplicate_correction: bool = True) -> th.Tensor:
    c_dot = th.empty_like(c)
    binding(A, B, c, c_dot, expand_dim, power, d_tile, scale_A, duplicate_correction) # type: ignore
    return c_dot

@op.register_fake
def op_fake(A: th.Tensor, B: th.Tensor, c: th.Tensor, expand_dim: int, power: int, d_tile: int, scale_A: float = 1.0, duplicate_correction: bool = True):
    return th.empty_like(c)


from functools import wraps
import torch
from unittest.mock import patch, Mock, MagicMock
from vidrial.jit import settings
from vidrial.jit.binding import make_binding, tune_and_update, PickBest, PickAny
from vidrial.jit.tuner import get_fn_name
from vidrial.jit.timingcache import ConfigTimingCache
from vidrial.jit.compiler import compile
import time

def add_one(x: torch.Tensor, sleep: float):
    if sleep == 0.:
        raise ValueError("I need some sleep!")
    time.sleep(sleep)
    return x + 1, sleep
def add_one_hash(kwargs):
    x = kwargs['x']
    hash = str(x.shape) + str(x.stride()) 
    hash += str(x.dtype) + str(x.device)
    return hash
add_one_sweep = [{'sleep': 0.}, {'sleep': 0.02}, {'sleep': 0.01}]

def test_policy_pickany(tmp_path):
    with patch('vidrial.jit.binding.tune_and_update', wraps=tune_and_update) as mock_tune_and_update:
        timing_cache = ConfigTimingCache(get_fn_name(add_one), add_one_hash, root=str(tmp_path))
        fn = make_binding(cache=timing_cache, sweep=add_one_sweep)(add_one)
        with settings.set(PickAny, 1):
            x = torch.randn(64)
            y, sleep = fn(x) # Should run with the first version in the sweep
            assert mock_tune_and_update.call_count == 0 # ensure it doesn't tune anything
            assert torch.allclose(y, x + 1)
            assert sleep != 0. # ensure it didn't pick the version that crashes

def test_policy_pickbest(tmp_path):
    with patch('vidrial.jit.binding.tune_and_update', wraps=tune_and_update) as mock_tune_and_update:
        timing_cache = ConfigTimingCache(get_fn_name(add_one), add_one_hash, root=str(tmp_path))
        fn = make_binding(cache=timing_cache, sweep=add_one_sweep)(add_one)
        with settings.set(PickBest, 1):
            x = torch.randn(64)
            fn(x) # First call should tune since cache is new
            assert mock_tune_and_update.call_count == 1
            mock_tune_and_update.reset_mock()
            y, sleep = fn(x) # call again, should have cache hit
            assert mock_tune_and_update.call_count == 0
            assert torch.allclose(y, x + 1)
            assert sleep == 0.01 # ensure it picks the best version that doesn't crash

def test_policy_pickbest_call_sweep(tmp_path):
    timing_cache = ConfigTimingCache(get_fn_name(add_one), add_one_hash, root=str(tmp_path))
    sweep_fn = lambda args: [{'sleep': 0.01}, {'sleep': 0.02}]
    mock_sweep_fn = Mock(side_effect=sweep_fn)
    fn = make_binding(cache=timing_cache, sweep=mock_sweep_fn)(add_one)
    with settings.set(PickBest, 1):
        x = torch.randn(64)
        fn(x) # First call should tune since cache is new
        assert mock_sweep_fn.call_count == 1
        fn(x) # call again, should not call the sweep because there is a cache hit
        assert mock_sweep_fn.call_count == 1

def test_bypass(tmp_path):
    timing_cache = ConfigTimingCache(get_fn_name(add_one), add_one_hash, root=str(tmp_path))
    sweep_fn = lambda args: [{'sleep': 0.01}, {'sleep': 0.02}]
    mock_sweep_fn = Mock(side_effect=sweep_fn)
    mock_call_tracker = []
    @wraps(add_one)
    def mock_add_one(*args, **kwargs):
        nonlocal mock_call_tracker
        mock_call_tracker.append((args, kwargs))
        return add_one(*args, **kwargs)
    fn = make_binding(cache=timing_cache, sweep=mock_sweep_fn)(mock_add_one)
    x = torch.randn(64)
    fn(x, sleep=0.01) # Call with all the arguments which should bypass any sweep call
    assert mock_sweep_fn.call_count == 0
    assert len(mock_call_tracker) == 1
    assert mock_call_tracker[0][1]['sleep'] == 0.01
